import decimal
import json
import io
import pytest

from werkzeug.exceptions import BadRequest
from werkzeug.wrappers import Request
from werkzeug.datastructures import FileStorage, MultiDict

from flask_restx import Api, Model, fields, inputs
from flask_restx.errors import SpecsError
from flask_restx.reqparse import Argument, RequestParser, ParseResult


class ReqParseTest(object):
    def test_api_shortcut(self, app):
        api = Api(app)
        parser = api.parser()
        assert isinstance(parser, RequestParser)

    def test_parse_model(self, app):
        model = Model("Todo", {"task": fields.String(required=True)})

        parser = RequestParser()
        parser.add_argument("todo", type=model, required=True)

        data = {"todo": {"task": "aaa"}}

        with app.test_request_context(
            "/", method="post", data=json.dumps(data), content_type="application/json"
        ):
            args = parser.parse_args()
            assert args["todo"] == {"task": "aaa"}

    def test_help(self, app, mocker):
        abort = mocker.patch(
            "flask_restx.reqparse.abort", side_effect=BadRequest("Bad Request")
        )
        parser = RequestParser()
        parser.add_argument("foo", choices=("one", "two"), help="Bad choice.")
        req = mocker.Mock(["values", "get_json"])
        req.values = MultiDict([("foo", "three")])
        req.get_json.return_value = None
        with pytest.raises(BadRequest):
            parser.parse_args(req)
        expected = {
            "foo": "Bad choice. The value 'three' is not a valid choice for 'foo'."
        }
        abort.assert_called_with(
            400, "Input payload validation failed", errors=expected
        )

    def test_no_help(self, app, mocker):
        abort = mocker.patch(
            "flask_restx.reqparse.abort", side_effect=BadRequest("Bad Request")
        )
        parser = RequestParser()
        parser.add_argument("foo", choices=["one", "two"])
        req = mocker.Mock(["values", "get_json"])
        req.get_json.return_value = None
        req.values = MultiDict([("foo", "three")])
        with pytest.raises(BadRequest):
            parser.parse_args(req)
        expected = {"foo": "The value 'three' is not a valid choice for 'foo'."}
        abort.assert_called_with(
            400, "Input payload validation failed", errors=expected
        )

    @pytest.mark.request_context()
    def test_viewargs(self, mocker):
        req = Request.from_values()
        req.view_args = {"foo": "bar"}
        parser = RequestParser()
        parser.add_argument("foo", location=["view_args"])
        args = parser.parse_args(req)
        assert args["foo"] == "bar"

        req = mocker.Mock(["get_json"])
        req.values = ()
        req.get_json.return_value = None
        req.view_args = {"foo": "bar"}
        parser = RequestParser()
        parser.add_argument("foo", store_missing=True)
        args = parser.parse_args(req)
        assert args["foo"] is None

    def test_parse_unicode(self, app):
        req = Request.from_values("/bubble?foo=barß")
        parser = RequestParser()
        parser.add_argument("foo")

        args = parser.parse_args(req)
        assert args["foo"] == "barß"

    def test_parse_unicode_app(self, app):
        parser = RequestParser()
        parser.add_argument("foo")

        with app.test_request_context("/bubble?foo=barß"):
            args = parser.parse_args()
            assert args["foo"] == "barß"

    @pytest.mark.request_context(
        "/bubble", method="post", content_type="application/json"
    )
    def test_json_location(self):
        parser = RequestParser()
        parser.add_argument("foo", location="json", store_missing=True)
        args = parser.parse_args()
        assert args["foo"] is None

    @pytest.mark.request_context(
        "/bubble",
        method="post",
        data=json.dumps({"foo": "bar"}),
        content_type="application/json",
    )
    def test_get_json_location(self):
        parser = RequestParser()
        parser.add_argument("foo", location="json")
        args = parser.parse_args()
        assert args["foo"] == "bar"

    @pytest.mark.request_context("/bubble?foo=bar")
    def test_parse_append_ignore(self, app):
        parser = RequestParser()
        parser.add_argument(
            "foo", ignore=True, type=int, action="append", store_missing=True
        ),

        args = parser.parse_args()
        assert args["foo"] is None

    @pytest.mark.request_context("/bubble?")
    def test_parse_append_default(self):
        parser = RequestParser()
        parser.add_argument("foo", action="append", store_missing=True),

        args = parser.parse_args()
        assert args["foo"] is None

    @pytest.mark.request_context("/bubble?foo=bar&foo=bat")
    def test_parse_append(self):
        parser = RequestParser()
        parser.add_argument("foo", action="append"),

        args = parser.parse_args()
        assert args["foo"] == ["bar", "bat"]

    @pytest.mark.request_context("/bubble?foo=bar")
    def test_parse_append_single(self):
        parser = RequestParser()
        parser.add_argument("foo", action="append"),

        args = parser.parse_args()
        assert args["foo"] == ["bar"]

    @pytest.mark.request_context("/bubble?foo=bar")
    def test_split_single(self):
        parser = RequestParser()
        parser.add_argument("foo", action="split"),

        args = parser.parse_args()
        assert args["foo"] == ["bar"]

    @pytest.mark.request_context("/bubble?foo=bar,bat")
    def test_split_multiple(self):
        parser = RequestParser()
        parser.add_argument("foo", action="split"),

        args = parser.parse_args()
        assert args["foo"] == ["bar", "bat"]

    @pytest.mark.request_context("/bubble?foo=1,2,3")
    def test_split_multiple_cast(self):
        parser = RequestParser()
        parser.add_argument("foo", type=int, action="split")

        args = parser.parse_args()
        assert args["foo"] == [1, 2, 3]

    @pytest.mark.request_context("/bubble?foo=bar")
    def test_parse_dest(self):
        parser = RequestParser()
        parser.add_argument("foo", dest="bat")

        args = parser.parse_args()
        assert args["bat"] == "bar"

    @pytest.mark.request_context("/bubble?foo>=bar&foo<=bat&foo=foo")
    def test_parse_gte_lte_eq(self):
        parser = RequestParser()
        parser.add_argument("foo", operators=[">=", "<=", "="], action="append"),

        args = parser.parse_args()
        assert args["foo"] == ["bar", "bat", "foo"]

    @pytest.mark.request_context("/bubble?foo>=bar")
    def test_parse_gte(self):
        parser = RequestParser()
        parser.add_argument("foo", operators=[">="])

        args = parser.parse_args()
        assert args["foo"] == "bar"

    @pytest.mark.request_context("/bubble?foo=bar")
    def test_parse_foo_operators_four_hunderd(self):
        parser = RequestParser()
        parser.add_argument("foo", type=int),
        with pytest.raises(BadRequest):
            parser.parse_args()

    @pytest.mark.request_context("/bubble")
    def test_parse_foo_operators_ignore(self):
        parser = RequestParser()
        parser.add_argument("foo", ignore=True, store_missing=True)

        args = parser.parse_args()
        assert args["foo"] is None

    @pytest.mark.request_context("/bubble?foo<=bar")
    def test_parse_lte_gte_mock(self, mocker):
        mock_type = mocker.Mock()

        parser = RequestParser()
        parser.add_argument("foo", type=mock_type, operators=["<="])

        parser.parse_args()
        mock_type.assert_called_with("bar", "foo", "<=")

    @pytest.mark.request_context("/bubble?foo<=bar")
    def test_parse_lte_gte_append(self):
        parser = RequestParser()
        parser.add_argument("foo", operators=["<=", "="], action="append")

        args = parser.parse_args()
        assert args["foo"] == ["bar"]

    @pytest.mark.request_context("/bubble?foo<=bar")
    def test_parse_lte_gte_missing(self):
        parser = RequestParser()
        parser.add_argument("foo", operators=["<=", "="])
        args = parser.parse_args()
        assert args["foo"] == "bar"

    @pytest.mark.request_context("/bubble?foo=bar&foo=bat")
    def test_parse_eq_other(self):
        parser = RequestParser()
        parser.add_argument("foo"),
        args = parser.parse_args()
        assert args["foo"] == "bar"

    @pytest.mark.request_context("/bubble?foo=bar")
    def test_parse_eq(self):
        parser = RequestParser()
        parser.add_argument("foo"),
        args = parser.parse_args()
        assert args["foo"] == "bar"

    @pytest.mark.request_context("/bubble?foo<=bar")
    def test_parse_lte(self):
        parser = RequestParser()
        parser.add_argument("foo", operators=["<="])

        args = parser.parse_args()
        assert args["foo"] == "bar"

    @pytest.mark.request_context("/bubble")
    def test_parse_required(self, app):
        parser = RequestParser()
        parser.add_argument("foo", required=True, location="values")

        expected = {
            "foo": "Missing required parameter in the post body or the query string"
        }
        with pytest.raises(BadRequest) as cm:
            parser.parse_args()

        assert cm.value.data["message"] == "Input payload validation failed"
        assert cm.value.data["errors"] == expected

        parser = RequestParser()
        parser.add_argument("bar", required=True, location=["values", "cookies"])

        expected = {
            "bar": (
                "Missing required parameter in the post body or the query "
                "string or the request's cookies"
            )
        }

        with pytest.raises(BadRequest) as cm:
            parser.parse_args()
        assert cm.value.data["message"] == "Input payload validation failed"
        assert cm.value.data["errors"] == expected

    @pytest.mark.request_context("/bubble")
    @pytest.mark.options(bundle_errors=True)
    def test_parse_error_bundling(self, app):
        parser = RequestParser()
        parser.add_argument("foo", required=True, location="values")
        parser.add_argument("bar", required=True, location=["values", "cookies"])

        with pytest.raises(BadRequest) as cm:
            parser.parse_args()

        assert cm.value.data["message"] == "Input payload validation failed"
        assert cm.value.data["errors"] == {
            "foo": "Missing required parameter in the post body or the query string",
            "bar": (
                "Missing required parameter in the post body or the query string "
                "or the request's cookies"
            ),
        }

    @pytest.mark.request_context("/bubble")
    @pytest.mark.options(bundle_errors=False)
    def test_parse_error_bundling_w_parser_arg(self, app):
        parser = RequestParser(bundle_errors=True)
        parser.add_argument("foo", required=True, location="values")
        parser.add_argument("bar", required=True, location=["values", "cookies"])

        with pytest.raises(BadRequest) as cm:
            parser.parse_args()

        assert cm.value.data["message"] == "Input payload validation failed"
        assert cm.value.data["errors"] == {
            "foo": "Missing required parameter in the post body or the query string",
            "bar": (
                "Missing required parameter in the post body or the query string "
                "or the request's cookies"
            ),
        }

    @pytest.mark.request_context("/bubble")
    def test_parse_default_append(self):
        parser = RequestParser()
        parser.add_argument("foo", default="bar", action="append", store_missing=True)

        args = parser.parse_args()

        assert args["foo"] == "bar"

    @pytest.mark.request_context("/bubble")
    def test_parse_default(self):
        parser = RequestParser()
        parser.add_argument("foo", default="bar", store_missing=True)

        args = parser.parse_args()
        assert args["foo"] == "bar"

    @pytest.mark.request_context("/bubble")
    def test_parse_callable_default(self):
        parser = RequestParser()
        parser.add_argument("foo", default=lambda: "bar", store_missing=True)

        args = parser.parse_args()
        assert args["foo"] == "bar"

    @pytest.mark.request_context("/bubble?foo=bar")
    def test_parse(self):
        parser = RequestParser()
        parser.add_argument("foo"),

        args = parser.parse_args()
        assert args["foo"] == "bar"

    @pytest.mark.request_context("/bubble")
    def test_parse_none(self):
        parser = RequestParser()
        parser.add_argument("foo")

        args = parser.parse_args()
        assert args["foo"] is None

    def test_parse_store_missing(self, app):
        req = Request.from_values("/bubble")

        parser = RequestParser()
        parser.add_argument("foo", store_missing=False)

        args = parser.parse_args(req)
        assert "foo" not in args

    def test_parse_choices_correct(self, app):
        req = Request.from_values("/bubble?foo=bat")

        parser = RequestParser()
        parser.add_argument("foo", choices=["bat"]),

        args = parser.parse_args(req)
        assert args["foo"] == "bat"

    def test_parse_choices(self, app):
        req = Request.from_values("/bubble?foo=bar")

        parser = RequestParser()
        parser.add_argument("foo", choices=["bat"]),

        with pytest.raises(BadRequest):
            parser.parse_args(req)

    def test_parse_choices_sensitive(self, app):
        req = Request.from_values("/bubble?foo=BAT")

        parser = RequestParser()
        parser.add_argument("foo", choices=["bat"], case_sensitive=True),

        with pytest.raises(BadRequest):
            parser.parse_args(req)

    def test_parse_choices_insensitive(self, app):
        req = Request.from_values("/bubble?foo=BAT")

        parser = RequestParser()
        parser.add_argument("foo", choices=["bat"], case_sensitive=False),

        args = parser.parse_args(req)
        assert "bat" == args.get("foo")

        # both choices and args are case_insensitive
        req = Request.from_values("/bubble?foo=bat")

        parser = RequestParser()
        parser.add_argument("foo", choices=["BAT"], case_sensitive=False),

        args = parser.parse_args(req)
        assert "bat" == args.get("foo")

    def test_parse_ignore(self, app):
        req = Request.from_values("/bubble?foo=bar")

        parser = RequestParser()
        parser.add_argument("foo", type=int, ignore=True, store_missing=True),

        args = parser.parse_args(req)
        assert args["foo"] is None

    def test_chaining(self):
        parser = RequestParser()
        assert parser is parser.add_argument("foo")

    def test_result_existence(self):
        result = ParseResult()
        result.foo = "bar"
        result["bar"] = "baz"
        assert result["foo"] == "bar"
        assert result.bar == "baz"

    def test_result_missing(self):
        result = ParseResult()
        pytest.raises(AttributeError, lambda: result.spam)
        pytest.raises(KeyError, lambda: result["eggs"])

    def test_result_configurability(self):
        req = Request.from_values()
        assert isinstance(RequestParser().parse_args(req), ParseResult)
        assert type(RequestParser(result_class=dict).parse_args(req)) is dict

    def test_none_argument(self, app):
        parser = RequestParser()
        parser.add_argument("foo", location="json")
        with app.test_request_context(
            "/bubble",
            method="post",
            data=json.dumps({"foo": None}),
            content_type="application/json",
        ):
            args = parser.parse_args()
        assert args["foo"] is None

    def test_type_callable(self, app):
        req = Request.from_values("/bubble?foo=1")

        parser = RequestParser()
        parser.add_argument("foo", type=lambda x: x, required=False),

        args = parser.parse_args(req)
        assert args["foo"] == "1"

    def test_type_callable_none(self, app):
        parser = RequestParser()
        parser.add_argument("foo", type=lambda x: x, location="json", required=False),

        with app.test_request_context(
            "/bubble",
            method="post",
            data=json.dumps({"foo": None}),
            content_type="application/json",
        ):
            args = parser.parse_args()
            assert args["foo"] is None

    def test_type_decimal(self, app):
        parser = RequestParser()
        parser.add_argument("foo", type=decimal.Decimal, location="json")

        with app.test_request_context(
            "/bubble",
            method="post",
            data=json.dumps({"foo": "1.0025"}),
            content_type="application/json",
        ):
            args = parser.parse_args()
            assert args["foo"] == decimal.Decimal("1.0025")

    def test_type_filestorage(self, app):
        parser = RequestParser()
        parser.add_argument("foo", type=FileStorage, location="files")

        fdata = "foo bar baz qux".encode("utf-8")
        with app.test_request_context(
            "/bubble", method="POST", data={"foo": (io.BytesIO(fdata), "baz.txt")}
        ):
            args = parser.parse_args()

            assert args["foo"].name == "foo"
            assert args["foo"].filename == "baz.txt"
            assert args["foo"].read() == fdata

    def test_filestorage_custom_type(self, app):
        def _custom_type(f):
            return FileStorage(
                stream=f.stream,
                filename="{0}aaaa".format(f.filename),
                name="{0}aaaa".format(f.name),
            )

        parser = RequestParser()
        parser.add_argument("foo", type=_custom_type, location="files")

        fdata = "foo bar baz qux".encode("utf-8")
        with app.test_request_context(
            "/bubble", method="POST", data={"foo": (io.BytesIO(fdata), "baz.txt")}
        ):
            args = parser.parse_args()

            assert args["foo"].name == "fooaaaa"
            assert args["foo"].filename == "baz.txtaaaa"
            assert args["foo"].read() == fdata

    def test_passing_arguments_object(self, app):
        req = Request.from_values("/bubble?foo=bar")
        parser = RequestParser()
        parser.add_argument(Argument("foo"))

        args = parser.parse_args(req)
        assert args["foo"] == "bar"

    def test_int_choice_types(self, app):
        parser = RequestParser()
        parser.add_argument("foo", type=int, choices=[1, 2, 3], location="json")

        with app.test_request_context(
            "/bubble",
            method="post",
            data=json.dumps({"foo": 5}),
            content_type="application/json",
        ):
            with pytest.raises(BadRequest):
                parser.parse_args()

    def test_int_range_choice_types(self, app):
        parser = RequestParser()
        parser.add_argument("foo", type=int, choices=range(100), location="json")

        with app.test_request_context(
            "/bubble",
            method="post",
            data=json.dumps({"foo": 101}),
            content_type="application/json",
        ):
            with pytest.raises(BadRequest):
                parser.parse_args()

    def test_request_parser_copy(self, app):
        req = Request.from_values("/bubble?foo=101&bar=baz")
        parser = RequestParser()
        foo_arg = Argument("foo", type=int)
        parser.args.append(foo_arg)
        parser_copy = parser.copy()

        # Deepcopy should create a clone of the argument object instead of
        # copying a reference to the new args list
        assert foo_arg not in parser_copy.args

        # Args added to new parser should not be added to the original
        bar_arg = Argument("bar")
        parser_copy.args.append(bar_arg)
        assert bar_arg not in parser.args

        args = parser_copy.parse_args(req)
        assert args["foo"] == 101
        assert args["bar"] == "baz"

    def test_request_parse_copy_including_settings(self):
        parser = RequestParser(trim=True, bundle_errors=True)
        parser_copy = parser.copy()

        assert parser.trim == parser_copy.trim
        assert parser.bundle_errors == parser_copy.bundle_errors

    def test_request_parser_replace_argument(self, app):
        req = Request.from_values("/bubble?foo=baz")
        parser = RequestParser()
        parser.add_argument("foo", type=int)
        parser_copy = parser.copy()
        parser_copy.replace_argument("foo")

        args = parser_copy.parse_args(req)
        assert args["foo"] == "baz"

    def test_both_json_and_values_location(self, app):
        parser = RequestParser()
        parser.add_argument("foo", type=int)
        parser.add_argument("baz", type=int)
        with app.test_request_context(
            "/bubble?foo=1",
            method="post",
            data=json.dumps({"baz": 2}),
            content_type="application/json",
        ):
            args = parser.parse_args()
            assert args["foo"] == 1
            assert args["baz"] == 2

    def test_not_json_location_and_content_type_json(self, app):
        parser = RequestParser()
        parser.add_argument("foo", location="args")

        with app.test_request_context(
            "/bubble", method="get", content_type="application/json"
        ):
            parser.parse_args()  # Should not raise a 400: BadRequest

    def test_request_parser_remove_argument(self):
        req = Request.from_values("/bubble?foo=baz")
        parser = RequestParser()
        parser.add_argument("foo", type=int)
        parser_copy = parser.copy()
        parser_copy.remove_argument("foo")

        args = parser_copy.parse_args(req)
        assert args == {}

    def test_strict_parsing_off(self):
        req = Request.from_values("/bubble?foo=baz")
        parser = RequestParser()
        args = parser.parse_args(req)
        assert args == {}

    def test_strict_parsing_on(self):
        req = Request.from_values("/bubble?foo=baz")
        parser = RequestParser()
        with pytest.raises(BadRequest):
            parser.parse_args(req, strict=True)

    def test_strict_parsing_off_partial_hit(self, app):
        req = Request.from_values("/bubble?foo=1&bar=bees&n=22")
        parser = RequestParser()
        parser.add_argument("foo", type=int)
        args = parser.parse_args(req)
        assert args["foo"] == 1

    def test_strict_parsing_on_partial_hit(self, app):
        req = Request.from_values("/bubble?foo=1&bar=bees&n=22")
        parser = RequestParser()
        parser.add_argument("foo", type=int)
        with pytest.raises(BadRequest):
            parser.parse_args(req, strict=True)

    def test_trim_argument(self, app):
        req = Request.from_values("/bubble?foo= 1 &bar=bees&n=22")
        parser = RequestParser()
        parser.add_argument("foo")
        args = parser.parse_args(req)
        assert args["foo"] == " 1 "

        parser = RequestParser()
        parser.add_argument("foo", trim=True)
        args = parser.parse_args(req)
        assert args["foo"] == "1"

        parser = RequestParser()
        parser.add_argument("foo", trim=True, type=int)
        args = parser.parse_args(req)
        assert args["foo"] == 1

    def test_trim_request_parser(self, app):
        req = Request.from_values("/bubble?foo= 1 &bar=bees&n=22")
        parser = RequestParser(trim=False)
        parser.add_argument("foo")
        args = parser.parse_args(req)
        assert args["foo"] == " 1 "

        parser = RequestParser(trim=True)
        parser.add_argument("foo")
        args = parser.parse_args(req)
        assert args["foo"] == "1"

        parser = RequestParser(trim=True)
        parser.add_argument("foo", type=int)
        args = parser.parse_args(req)
        assert args["foo"] == 1

    def test_trim_request_parser_override_by_argument(self):
        parser = RequestParser(trim=True)
        parser.add_argument("foo", trim=False)

        assert parser.args[0].trim is False

    def test_trim_request_parser_json(self, app):
        parser = RequestParser(trim=True)
        parser.add_argument("foo", location="json")
        parser.add_argument("int1", location="json", type=int)
        parser.add_argument("int2", location="json", type=int)

        with app.test_request_context(
            "/bubble",
            method="post",
            data=json.dumps({"foo": " bar ", "int1": 1, "int2": " 2 "}),
            content_type="application/json",
        ):
            args = parser.parse_args()
        assert args["foo"] == "bar"
        assert args["int1"] == 1
        assert args["int2"] == 2


class ArgumentTest(object):
    def test_name(self):
        arg = Argument("foo")
        assert arg.name == "foo"

    def test_dest(self):
        arg = Argument("foo", dest="foobar")
        assert arg.dest == "foobar"

    def test_location_url(self):
        arg = Argument("foo", location="url")
        assert arg.location == "url"

    def test_location_url_list(self):
        arg = Argument("foo", location=["url"])
        assert arg.location == ["url"]

    def test_location_header(self):
        arg = Argument("foo", location="headers")
        assert arg.location == "headers"

    def test_location_json(self):
        arg = Argument("foo", location="json")
        assert arg.location == "json"

    def test_location_get_json(self):
        arg = Argument("foo", location="get_json")
        assert arg.location == "get_json"

    def test_location_header_list(self):
        arg = Argument("foo", location=["headers"])
        assert arg.location == ["headers"]

    def test_type(self):
        arg = Argument("foo", type=int)
        assert arg.type == int

    def test_default(self):
        arg = Argument("foo", default=True)
        assert arg.default is True

    def test_default_help(self):
        arg = Argument("foo")
        assert arg.help is None

    def test_required(self):
        arg = Argument("foo", required=True)
        assert arg.required is True

    def test_ignore(self):
        arg = Argument("foo", ignore=True)
        assert arg.ignore is True

    def test_operator(self):
        arg = Argument("foo", operators=[">=", "<=", "="])
        assert arg.operators == [">=", "<=", "="]

    def test_action_filter(self):
        arg = Argument("foo", action="filter")
        assert arg.action == "filter"

    def test_action(self):
        arg = Argument("foo", action="append")
        assert arg.action == "append"

    def test_choices(self):
        arg = Argument("foo", choices=[1, 2])
        assert arg.choices == [1, 2]

    def test_default_dest(self):
        arg = Argument("foo")
        assert arg.dest is None

    def test_default_operators(self):
        arg = Argument("foo")
        assert arg.operators[0] == "="
        assert len(arg.operators) == 1

    def test_default_type(self):
        arg = Argument("foo")
        sentinel = 666
        assert arg.type(sentinel) == "666"

    def test_default_default(self):
        arg = Argument("foo")
        assert arg.default is None

    def test_required_default(self):
        arg = Argument("foo")
        assert arg.required is False

    def test_ignore_default(self):
        arg = Argument("foo")
        assert arg.ignore is False

    def test_action_default(self):
        arg = Argument("foo")
        assert arg.action == "store"

    def test_choices_default(self):
        arg = Argument("foo")
        assert len(arg.choices) == 0

    def test_source(self, mocker):
        req = mocker.Mock(["args", "headers", "values"])
        req.args = {"foo": "bar"}
        req.headers = {"baz": "bat"}
        arg = Argument("foo", location=["args"])
        assert arg.source(req) == MultiDict(req.args)

        arg = Argument("foo", location=["headers"])
        assert arg.source(req) == MultiDict(req.headers)

    def test_convert_default_type_with_null_input(self):
        arg = Argument("foo")
        assert arg.convert(None, None) is None

    def test_convert_with_null_input_when_not_nullable(self):
        arg = Argument("foo", nullable=False)
        pytest.raises(ValueError, lambda: arg.convert(None, None))

    def test_source_bad_location(self, mocker):
        req = mocker.Mock(["values"])
        arg = Argument("foo", location=["foo"])
        assert len(arg.source(req)) == 0  # yes, basically you don't find it

    def test_source_default_location(self, mocker):
        req = mocker.Mock(["values", "get_json"])
        req.get_json.return_value = None
        req._get_child_mock = lambda **kwargs: MultiDict()
        arg = Argument("foo")
        assert arg.source(req) == req.values

    def test_option_case_sensitive(self):
        arg = Argument("foo", choices=["bar", "baz"], case_sensitive=True)
        assert arg.case_sensitive is True

        # Insensitive
        arg = Argument("foo", choices=["bar", "baz"], case_sensitive=False)
        assert arg.case_sensitive is False

        # Default
        arg = Argument("foo", choices=["bar", "baz"])
        assert arg.case_sensitive is True


class RequestParserSchemaTest(object):
    def test_empty_parser(self):
        parser = RequestParser()
        assert parser.__schema__ == []

    def test_primitive_types(self):
        parser = RequestParser()
        parser.add_argument("int", type=int, help="Some integer")
        parser.add_argument("str", type=str, help="Some string")
        parser.add_argument("float", type=float, help="Some float")

        assert parser.__schema__ == [
            {
                "description": "Some integer",
                "type": "integer",
                "name": "int",
                "in": "query",
            },
            {
                "description": "Some string",
                "type": "string",
                "name": "str",
                "in": "query",
            },
            {
                "description": "Some float",
                "type": "number",
                "name": "float",
                "in": "query",
            },
        ]

    def test_unknown_type(self):
        parser = RequestParser()
        parser.add_argument("unknown", type=lambda v: v)
        assert parser.__schema__ == [
            {
                "name": "unknown",
                "type": "string",
                "in": "query",
            }
        ]

    def test_required(self):
        parser = RequestParser()
        parser.add_argument("int", type=int, required=True)
        assert parser.__schema__ == [
            {
                "name": "int",
                "type": "integer",
                "in": "query",
                "required": True,
            }
        ]

    def test_default(self):
        parser = RequestParser()
        parser.add_argument("int", type=int, default=5)
        assert parser.__schema__ == [
            {
                "name": "int",
                "type": "integer",
                "in": "query",
                "default": 5,
            }
        ]

    def test_default_as_false(self):
        parser = RequestParser()
        parser.add_argument("bool", type=inputs.boolean, default=False)
        assert parser.__schema__ == [
            {
                "name": "bool",
                "type": "boolean",
                "in": "query",
                "default": False,
            }
        ]

    def test_choices(self):
        parser = RequestParser()
        parser.add_argument("string", type=str, choices=["a", "b"])
        assert parser.__schema__ == [
            {
                "name": "string",
                "type": "string",
                "in": "query",
                "enum": ["a", "b"],
            }
        ]

    def test_location(self):
        parser = RequestParser()
        parser.add_argument("default", type=int)
        parser.add_argument("in_values", type=int, location="values")
        parser.add_argument("in_query", type=int, location="args")
        parser.add_argument("in_headers", type=int, location="headers")
        parser.add_argument("in_cookie", type=int, location="cookie")
        assert parser.__schema__ == [
            {
                "name": "default",
                "type": "integer",
                "in": "query",
            },
            {
                "name": "in_values",
                "type": "integer",
                "in": "query",
            },
            {
                "name": "in_query",
                "type": "integer",
                "in": "query",
            },
            {
                "name": "in_headers",
                "type": "integer",
                "in": "header",
            },
        ]

    def test_location_json(self):
        parser = RequestParser()
        parser.add_argument("in_json", type=str, location="json")
        assert parser.__schema__ == [
            {
                "name": "in_json",
                "type": "string",
                "in": "body",
            }
        ]

    def test_location_form(self):
        parser = RequestParser()
        parser.add_argument("in_form", type=int, location="form")
        assert parser.__schema__ == [
            {
                "name": "in_form",
                "type": "integer",
                "in": "formData",
            }
        ]

    def test_location_files(self):
        parser = RequestParser()
        parser.add_argument("in_files", type=FileStorage, location="files")
        assert parser.__schema__ == [
            {
                "name": "in_files",
                "type": "file",
                "in": "formData",
            }
        ]

    def test_form_and_body_location(self):
        parser = RequestParser()
        parser.add_argument("default", type=int)
        parser.add_argument("in_form", type=int, location="form")
        parser.add_argument("in_json", type=str, location="json")
        with pytest.raises(SpecsError) as cm:
            parser.__schema__

        assert cm.value.msg == "Can't use formData and body at the same time"

    def test_files_and_body_location(self):
        parser = RequestParser()
        parser.add_argument("default", type=int)
        parser.add_argument("in_files", type=FileStorage, location="files")
        parser.add_argument("in_json", type=str, location="json")
        with pytest.raises(SpecsError) as cm:
            parser.__schema__

        assert cm.value.msg == "Can't use formData and body at the same time"

    def test_models(self):
        todo_fields = Model(
            "Todo",
            {"task": fields.String(required=True, description="The task details")},
        )
        parser = RequestParser()
        parser.add_argument("todo", type=todo_fields)
        assert parser.__schema__ == [
            {
                "name": "todo",
                "type": "Todo",
                "in": "body",
            }
        ]

    def test_lists(self):
        parser = RequestParser()
        parser.add_argument("int", type=int, action="append")
        assert parser.__schema__ == [
            {
                "name": "int",
                "in": "query",
                "type": "array",
                "collectionFormat": "multi",
                "items": {"type": "integer"},
            }
        ]

    def test_split_lists(self):
        parser = RequestParser()
        parser.add_argument("int", type=int, action="split")
        assert parser.__schema__ == [
            {
                "name": "int",
                "in": "query",
                "type": "array",
                "collectionFormat": "csv",
                "items": {"type": "integer"},
            }
        ]

    def test_schema_interface(self):
        def custom(value):
            pass

        custom.__schema__ = {
            "type": "string",
            "format": "custom-format",
        }

        parser = RequestParser()
        parser.add_argument("custom", type=custom)

        assert parser.__schema__ == [
            {
                "name": "custom",
                "in": "query",
                "type": "string",
                "format": "custom-format",
            }
        ]

    def test_callable_default(self):
        parser = RequestParser()
        parser.add_argument("int", type=int, default=lambda: 5)
        assert parser.__schema__ == [
            {
                "name": "int",
                "type": "integer",
                "in": "query",
                "default": 5,
            }
        ]



import copy

from flask import url_for, Blueprint

import flask_restx as restx


class APITest(object):
    def test_root_endpoint(self, app):
        api = restx.Api(app, version="1.0")

        with app.test_request_context():
            url = url_for("root")
            assert url == "/"
            assert api.base_url == "http://localhost/"

    def test_root_endpoint_lazy(self, app):
        api = restx.Api(version="1.0")
        api.init_app(app)

        with app.test_request_context():
            url = url_for("root")
            assert url == "/"
            assert api.base_url == "http://localhost/"

    def test_root_endpoint_with_blueprint(self, app):
        blueprint = Blueprint("api", __name__, url_prefix="/api")
        api = restx.Api(blueprint, version="1.0")
        app.register_blueprint(blueprint)

        with app.test_request_context():
            url = url_for("api.root")
            assert url == "/api/"
            assert api.base_url == "http://localhost/api/"

    def test_root_endpoint_with_blueprint_with_subdomain(self, app):
        blueprint = Blueprint("api", __name__, subdomain="api", url_prefix="/api")
        api = restx.Api(blueprint, version="1.0")
        app.register_blueprint(blueprint)

        with app.test_request_context():
            url = url_for("api.root")
            assert url == "http://api.localhost/api/"
            assert api.base_url == "http://api.localhost/api/"

    def test_parser(self):
        api = restx.Api()
        assert isinstance(api.parser(), restx.reqparse.RequestParser)

    def test_doc_decorator(self, app):
        api = restx.Api(app, prefix="/api", version="1.0")
        params = {"q": {"description": "some description"}}

        @api.doc(params=params)
        class TestResource(restx.Resource):
            pass

        assert hasattr(TestResource, "__apidoc__")
        assert TestResource.__apidoc__ == {"params": params}

    def test_doc_with_inheritance(self, app):
        api = restx.Api(app, prefix="/api", version="1.0")
        base_params = {
            "q": {
                "description": "some description",
                "type": "string",
                "paramType": "query",
            }
        }
        child_params = {
            "q": {"description": "some new description"},
            "other": {"description": "another param"},
        }

        @api.doc(params=base_params)
        class BaseResource(restx.Resource):
            pass

        @api.doc(params=child_params)
        class TestResource(BaseResource):
            pass

        assert TestResource.__apidoc__ == {
            "params": {
                "q": {
                    "description": "some new description",
                    "type": "string",
                    "paramType": "query",
                },
                "other": {"description": "another param"},
            }
        }

    def test_specs_endpoint_not_added(self, app):
        api = restx.Api()
        api.init_app(app, add_specs=False)
        assert "specs" not in api.endpoints
        assert "specs" not in app.view_functions

    def test_specs_endpoint_not_found_if_not_added(self, app, client):
        api = restx.Api()
        api.init_app(app, add_specs=False)
        resp = client.get("/swagger.json")
        assert resp.status_code == 404

    def test_default_endpoint(self, app):
        api = restx.Api(app)

        @api.route("/test/")
        class TestResource(restx.Resource):
            pass

        with app.test_request_context():
            assert url_for("test_resource") == "/test/"

    def test_default_endpoint_lazy(self, app):
        api = restx.Api()

        @api.route("/test/")
        class TestResource(restx.Resource):
            pass

        api.init_app(app)

        with app.test_request_context():
            assert url_for("test_resource") == "/test/"

    def test_default_endpoint_with_blueprint(self, app):
        blueprint = Blueprint("api", __name__, url_prefix="/api")
        api = restx.Api(blueprint)
        app.register_blueprint(blueprint)

        @api.route("/test/")
        class TestResource(restx.Resource):
            pass

        with app.test_request_context():
            assert url_for("api.test_resource") == "/api/test/"

    def test_default_endpoint_with_blueprint_with_subdomain(self, app):
        blueprint = Blueprint("api", __name__, subdomain="api", url_prefix="/api")
        api = restx.Api(blueprint)
        app.register_blueprint(blueprint)

        @api.route("/test/")
        class TestResource(restx.Resource):
            pass

        with app.test_request_context():
            assert url_for("api.test_resource") == "http://api.localhost/api/test/"

    def test_default_endpoint_for_namespace(self, app):
        api = restx.Api(app)
        ns = api.namespace("ns", "Test namespace")

        @ns.route("/test/")
        class TestResource(restx.Resource):
            pass

        with app.test_request_context():
            assert url_for("ns_test_resource") == "/ns/test/"

    def test_default_endpoint_lazy_for_namespace(self, app):
        api = restx.Api()
        ns = api.namespace("ns", "Test namespace")

        @ns.route("/test/")
        class TestResource(restx.Resource):
            pass

        api.init_app(app)

        with app.test_request_context():
            assert url_for("ns_test_resource") == "/ns/test/"

    def test_default_endpoint_for_namespace_with_blueprint(self, app):
        blueprint = Blueprint("api", __name__, url_prefix="/api")
        api = restx.Api(blueprint)
        ns = api.namespace("ns", "Test namespace")

        @ns.route("/test/")
        class TestResource(restx.Resource):
            pass

        app.register_blueprint(blueprint)

        with app.test_request_context():
            assert url_for("api.ns_test_resource") == "/api/ns/test/"

    def test_multiple_default_endpoint(self, app):
        api = restx.Api(app)

        @api.route("/test2/")
        @api.route("/test/")
        class TestResource(restx.Resource):
            pass

        with app.test_request_context():
            assert url_for("test_resource") == "/test/"
            assert url_for("test_resource_2") == "/test2/"

    def test_multiple_default_endpoint_lazy(self, app):
        api = restx.Api()

        @api.route("/test2/")
        @api.route("/test/")
        class TestResource(restx.Resource):
            pass

        api.init_app(app)

        with app.test_request_context():
            assert url_for("test_resource") == "/test/"
            assert url_for("test_resource_2") == "/test2/"

    def test_multiple_default_endpoint_for_namespace(self, app):
        api = restx.Api(app)
        ns = api.namespace("ns", "Test namespace")

        @ns.route("/test2/")
        @ns.route("/test/")
        class TestResource(restx.Resource):
            pass

        with app.test_request_context():
            assert url_for("ns_test_resource") == "/ns/test/"
            assert url_for("ns_test_resource_2") == "/ns/test2/"

    def test_multiple_default_endpoint_lazy_for_namespace(self, app):
        api = restx.Api()
        ns = api.namespace("ns", "Test namespace")

        @ns.route("/test2/")
        @ns.route("/test/")
        class TestResource(restx.Resource):
            pass

        api.init_app(app)

        with app.test_request_context():
            assert url_for("ns_test_resource") == "/ns/test/"
            assert url_for("ns_test_resource_2") == "/ns/test2/"

    def test_multiple_default_endpoint_for_namespace_with_blueprint(self, app):
        blueprint = Blueprint("api", __name__, url_prefix="/api")
        api = restx.Api(blueprint)
        ns = api.namespace("ns", "Test namespace")

        @ns.route("/test2/")
        @ns.route("/test/")
        class TestResource(restx.Resource):
            pass

        app.register_blueprint(blueprint)

        with app.test_request_context():
            assert url_for("api.ns_test_resource") == "/api/ns/test/"
            assert url_for("api.ns_test_resource_2") == "/api/ns/test2/"

    def test_ns_path_prefixes(self, app):
        api = restx.Api()
        ns = restx.Namespace("test_ns", description="Test namespace")

        @ns.route("/test/", endpoint="test_resource")
        class TestResource(restx.Resource):
            pass

        api.add_namespace(ns, "/api_test")
        api.init_app(app)

        with app.test_request_context():
            assert url_for("test_resource") == "/api_test/test/"

    def test_multiple_ns_with_authorizations(self, app):
        api = restx.Api()
        a1 = {"apikey": {"type": "apiKey", "in": "header", "name": "X-API"}}
        a2 = {
            "oauth2": {
                "type": "oauth2",
                "flow": "accessCode",
                "tokenUrl": "https://somewhere.com/token",
                "scopes": {
                    "read": "Grant read-only access",
                    "write": "Grant read-write access",
                },
            }
        }
        ns1 = restx.Namespace("ns1", authorizations=a1)
        ns2 = restx.Namespace("ns2", authorizations=a2)

        @ns1.route("/")
        class Ns1(restx.Resource):
            @ns1.doc(security="apikey")
            def get(self):
                pass

        @ns2.route("/")
        class Ns2(restx.Resource):
            @ns1.doc(security="oauth2")
            def post(self):
                pass

        api.add_namespace(ns1, path="/ns1")
        api.add_namespace(ns2, path="/ns2")
        api.init_app(app)

        assert {"apikey": []} in api.__schema__["paths"]["/ns1/"]["get"]["security"]
        assert {"oauth2": []} in api.__schema__["paths"]["/ns2/"]["post"]["security"]
        unified_auth = copy.copy(a1)
        unified_auth.update(a2)
        assert api.__schema__["securityDefinitions"] == unified_auth

    def test_non_ordered_namespace(self, app):
        api = restx.Api(app)
        ns = api.namespace("ns", "Test namespace")

        assert not ns.ordered

    def test_ordered_namespace(self, app):
        api = restx.Api(app, ordered=True)
        ns = api.namespace("ns", "Test namespace")

        assert ns.ordered

    def test_decorators(self, app, mocker):
        decorator1 = mocker.Mock(return_value=lambda x: x)
        decorator2 = mocker.Mock(return_value=lambda x: x)
        decorator3 = mocker.Mock(return_value=lambda x: x)

        class TestResource(restx.Resource):
            method_decorators = []

        api = restx.Api(decorators=[decorator1])
        ns = api.namespace("test_ns", decorators=[decorator2, decorator3])

        ns.add_resource(TestResource, "/test", endpoint="test")
        api.init_app(app)

        assert decorator1.called is True
        assert decorator2.called is True
        assert decorator3.called is True

    def test_specs_url(self, app):
        api = restx.Api(app)
        specs_url = api.specs_url
        assert specs_url == "/swagger.json"

    def test_url_scheme(self, app):
        api = restx.Api(app, url_scheme="https")
        assert api.specs_url == "https://localhost/swagger.json"
        assert api.base_url == "https://localhost/"



from flask_restx import Api, Resource, cors


class ErrorsTest(object):
    def test_crossdomain(self, app, client):
        class Foo(Resource):
            @cors.crossdomain(origin="*")
            def get(self):
                return "data"

        api = Api(app)
        api.add_resource(Foo, "/test/")

        res = client.get("/test/")
        assert res.status_code == 200
        assert res.headers["Access-Control-Allow-Origin"] == "*"
        assert res.headers["Access-Control-Max-Age"] == "21600"
        assert "HEAD" in res.headers["Access-Control-Allow-Methods"]
        assert "OPTIONS" in res.headers["Access-Control-Allow-Methods"]
        assert "GET" in res.headers["Access-Control-Allow-Methods"]

    def test_access_control_expose_headers(self, app, client):
        class Foo(Resource):
            @cors.crossdomain(
                origin="*", expose_headers=["X-My-Header", "X-Another-Header"]
            )
            def get(self):
                return "data"

        api = Api(app)
        api.add_resource(Foo, "/test/")

        res = client.get("/test/")
        assert res.status_code == 200
        assert "X-MY-HEADER" in res.headers["Access-Control-Expose-Headers"]
        assert "X-ANOTHER-HEADER" in res.headers["Access-Control-Expose-Headers"]

    def test_no_crossdomain(self, app, client):
        class Foo(Resource):
            def get(self):
                return "data"

        api = Api(app)
        api.add_resource(Foo, "/test/")

        res = client.get("/test/")
        assert res.status_code == 200
        assert "Access-Control-Allow-Origin" not in res.headers
        assert "Access-Control-Allow-Methods" not in res.headers
        assert "Access-Control-Max-Age" not in res.headers



import json
import pytest

from collections import OrderedDict

from flask_restx import mask, Api, Resource, fields, marshal, Mask


def assert_data(tested, expected):
    """Compare data without caring about order and type (dict vs. OrderedDict)"""
    tested = json.loads(json.dumps(tested))
    expected = json.loads(json.dumps(expected))
    assert tested == expected


class MaskMixin(object):
    def test_empty_mask(self):
        assert Mask("") == {}

    def test_one_field(self):
        assert Mask("field_name") == {"field_name": True}

    def test_multiple_field(self):
        mask = Mask("field1, field2, field3")
        assert_data(
            mask,
            {
                "field1": True,
                "field2": True,
                "field3": True,
            },
        )

    def test_nested_fields(self):
        parsed = Mask("nested{field1,field2}")
        expected = {
            "nested": {
                "field1": True,
                "field2": True,
            }
        }
        assert parsed == expected

    def test_complex(self):
        parsed = Mask("field1, nested{field, sub{subfield}}, field2")
        expected = {
            "field1": True,
            "nested": {
                "field": True,
                "sub": {
                    "subfield": True,
                },
            },
            "field2": True,
        }
        assert_data(parsed, expected)

    def test_star(self):
        parsed = Mask("nested{field1,field2},*")
        expected = {
            "nested": {
                "field1": True,
                "field2": True,
            },
            "*": True,
        }
        assert_data(parsed, expected)

    def test_order(self):
        parsed = Mask("f_3, nested{f_1, f_2, f_3}, f_2, f_1")
        expected = OrderedDict(
            [
                ("f_3", True),
                (
                    "nested",
                    OrderedDict(
                        [
                            ("f_1", True),
                            ("f_2", True),
                            ("f_3", True),
                        ]
                    ),
                ),
                ("f_2", True),
                ("f_1", True),
            ]
        )
        assert parsed == expected

    def test_missing_closing_bracket(self):
        with pytest.raises(mask.ParseError):
            Mask("nested{")

    def test_consecutive_coma(self):
        with pytest.raises(mask.ParseError):
            Mask("field,,")

    def test_coma_before_bracket(self):
        with pytest.raises(mask.ParseError):
            Mask("field,{}")

    def test_coma_after_bracket(self):
        with pytest.raises(mask.ParseError):
            Mask("nested{,}")

    def test_unexpected_opening_bracket(self):
        with pytest.raises(mask.ParseError):
            Mask("{{field}}")

    def test_unexpected_closing_bracket(self):
        with pytest.raises(mask.ParseError):
            Mask("{field}}")

    def test_support_colons(self):
        assert Mask("field:name") == {"field:name": True}

    def test_support_dash(self):
        assert Mask("field-name") == {"field-name": True}

    def test_support_underscore(self):
        assert Mask("field_name") == {"field_name": True}


class MaskUnwrappedTest(MaskMixin):
    def parse(self, value):
        return Mask(value)


class MaskWrappedTest(MaskMixin):
    def parse(self, value):
        return Mask("{" + value + "}")


class DObject(object):
    """A dead simple object built from a dictionnary (no recursion)"""

    def __init__(self, data):
        self.__dict__.update(data)


person_fields = {"name": fields.String, "age": fields.Integer}


class ApplyMaskTest(object):
    def test_empty(self):
        data = {
            "integer": 42,
            "string": "a string",
            "boolean": True,
        }
        result = mask.apply(data, "{}")
        assert result == {}

    def test_single_field(self):
        data = {
            "integer": 42,
            "string": "a string",
            "boolean": True,
        }
        result = mask.apply(data, "{integer}")
        assert result == {"integer": 42}

    def test_multiple_fields(self):
        data = {
            "integer": 42,
            "string": "a string",
            "boolean": True,
        }
        result = mask.apply(data, "{integer, string}")
        assert result == {"integer": 42, "string": "a string"}

    def test_star_only(self):
        data = {
            "integer": 42,
            "string": "a string",
            "boolean": True,
        }
        result = mask.apply(data, "*")
        assert result == data

    def test_with_objects(self):
        data = DObject(
            {
                "integer": 42,
                "string": "a string",
                "boolean": True,
            }
        )
        result = mask.apply(data, "{integer, string}")
        assert result == {"integer": 42, "string": "a string"}

    def test_with_ordered_dict(self):
        data = OrderedDict(
            {
                "integer": 42,
                "string": "a string",
                "boolean": True,
            }
        )
        result = mask.apply(data, "{integer, string}")
        assert result == {"integer": 42, "string": "a string"}

    def test_nested_field(self):
        data = {
            "integer": 42,
            "string": "a string",
            "boolean": True,
            "nested": {
                "integer": 42,
                "string": "a string",
                "boolean": True,
            },
        }
        result = mask.apply(data, "{nested}")
        assert result == {
            "nested": {
                "integer": 42,
                "string": "a string",
                "boolean": True,
            }
        }

    def test_nested_fields(self):
        data = {
            "nested": {
                "integer": 42,
                "string": "a string",
                "boolean": True,
            }
        }
        result = mask.apply(data, "{nested{integer}}")
        assert result == {"nested": {"integer": 42}}

    def test_nested_with_start(self):
        data = {
            "nested": {
                "integer": 42,
                "string": "a string",
                "boolean": True,
            },
            "other": "value",
        }
        result = mask.apply(data, "{nested{integer},*}")
        assert result == {"nested": {"integer": 42}, "other": "value"}

    def test_nested_fields_when_none(self):
        data = {"nested": None}
        result = mask.apply(data, "{nested{integer}}")
        assert result == {"nested": None}

    def test_raw_api_fields(self):
        family_fields = {
            "father": fields.Raw,
            "mother": fields.Raw,
        }

        result = mask.apply(family_fields, "father{name},mother{age}")

        data = {
            "father": {"name": "John", "age": 42},
            "mother": {"name": "Jane", "age": 42},
        }
        expected = {"father": {"name": "John"}, "mother": {"age": 42}}

        assert_data(marshal(data, result), expected)
        # Should leave th original mask untouched
        assert_data(marshal(data, family_fields), data)

    def test_nested_api_fields(self):
        family_fields = {
            "father": fields.Nested(person_fields),
            "mother": fields.Nested(person_fields),
        }

        result = mask.apply(family_fields, "father{name},mother{age}")
        assert set(result.keys()) == set(["father", "mother"])
        assert isinstance(result["father"], fields.Nested)
        assert set(result["father"].nested.keys()) == set(["name"])
        assert isinstance(result["mother"], fields.Nested)
        assert set(result["mother"].nested.keys()) == set(["age"])

        data = {
            "father": {"name": "John", "age": 42},
            "mother": {"name": "Jane", "age": 42},
        }
        expected = {"father": {"name": "John"}, "mother": {"age": 42}}

        assert_data(marshal(data, result), expected)
        # Should leave th original mask untouched
        assert_data(marshal(data, family_fields), data)

    def test_multiple_nested_api_fields(self):
        level_2 = {"nested_2": fields.Nested(person_fields)}
        level_1 = {"nested_1": fields.Nested(level_2)}
        root = {"nested": fields.Nested(level_1)}

        result = mask.apply(root, "nested{nested_1{nested_2{name}}}")
        assert set(result.keys()) == set(["nested"])
        assert isinstance(result["nested"], fields.Nested)
        assert set(result["nested"].nested.keys()) == set(["nested_1"])

        data = {"nested": {"nested_1": {"nested_2": {"name": "John", "age": 42}}}}
        expected = {"nested": {"nested_1": {"nested_2": {"name": "John"}}}}

        assert_data(marshal(data, result), expected)
        # Should leave th original mask untouched
        assert_data(marshal(data, root), data)

    def test_list_fields_with_simple_field(self):
        family_fields = {"name": fields.String, "members": fields.List(fields.String)}

        result = mask.apply(family_fields, "members")
        assert set(result.keys()) == set(["members"])
        assert isinstance(result["members"], fields.List)
        assert isinstance(result["members"].container, fields.String)

        data = {"name": "Doe", "members": ["John", "Jane"]}
        expected = {"members": ["John", "Jane"]}

        assert_data(marshal(data, result), expected)
        # Should leave th original mask untouched
        assert_data(marshal(data, family_fields), data)

    def test_list_fields_with_nested(self):
        family_fields = {"members": fields.List(fields.Nested(person_fields))}

        result = mask.apply(family_fields, "members{name}")
        assert set(result.keys()) == set(["members"])
        assert isinstance(result["members"], fields.List)
        assert isinstance(result["members"].container, fields.Nested)
        assert set(result["members"].container.nested.keys()) == set(["name"])

        data = {
            "members": [
                {"name": "John", "age": 42},
                {"name": "Jane", "age": 42},
            ]
        }
        expected = {"members": [{"name": "John"}, {"name": "Jane"}]}

        assert_data(marshal(data, result), expected)
        # Should leave th original mask untouched
        assert_data(marshal(data, family_fields), data)

    def test_list_fields_with_nested_inherited(self, app):
        api = Api(app)

        person = api.model("Person", {"name": fields.String, "age": fields.Integer})
        child = api.inherit("Child", person, {"attr": fields.String})

        family = api.model("Family", {"children": fields.List(fields.Nested(child))})

        result = mask.apply(family.resolved, "children{name,attr}")

        data = {
            "children": [
                {"name": "John", "age": 5, "attr": "value-john"},
                {"name": "Jane", "age": 42, "attr": "value-jane"},
            ]
        }
        expected = {
            "children": [
                {"name": "John", "attr": "value-john"},
                {"name": "Jane", "attr": "value-jane"},
            ]
        }

        assert_data(marshal(data, result), expected)
        # Should leave th original mask untouched
        assert_data(marshal(data, family), data)

    def test_list_fields_with_raw(self):
        family_fields = {"members": fields.List(fields.Raw)}

        result = mask.apply(family_fields, "members{name}")

        data = {
            "members": [
                {"name": "John", "age": 42},
                {"name": "Jane", "age": 42},
            ]
        }
        expected = {"members": [{"name": "John"}, {"name": "Jane"}]}

        assert_data(marshal(data, result), expected)
        # Should leave th original mask untouched
        assert_data(marshal(data, family_fields), data)

    def test_list(self):
        data = [
            {
                "integer": 42,
                "string": "a string",
                "boolean": True,
            },
            {
                "integer": 404,
                "string": "another string",
                "boolean": False,
            },
        ]
        result = mask.apply(data, "{integer, string}")
        assert result == [
            {"integer": 42, "string": "a string"},
            {"integer": 404, "string": "another string"},
        ]

    def test_nested_list(self):
        data = {
            "integer": 42,
            "list": [
                {
                    "integer": 42,
                    "string": "a string",
                },
                {
                    "integer": 404,
                    "string": "another string",
                },
            ],
        }
        result = mask.apply(data, "{list}")
        assert result == {
            "list": [
                {
                    "integer": 42,
                    "string": "a string",
                },
                {
                    "integer": 404,
                    "string": "another string",
                },
            ]
        }

    def test_nested_list_fields(self):
        data = {
            "list": [
                {
                    "integer": 42,
                    "string": "a string",
                },
                {
                    "integer": 404,
                    "string": "another string",
                },
            ]
        }
        result = mask.apply(data, "{list{integer}}")
        assert result == {"list": [{"integer": 42}, {"integer": 404}]}

    def test_missing_field_none_by_default(self):
        result = mask.apply({}, "{integer}")
        assert result == {"integer": None}

    def test_missing_field_skipped(self):
        result = mask.apply({}, "{integer}", skip=True)
        assert result == {}

    def test_missing_nested_field_skipped(self):
        result = mask.apply({}, "nested{integer}", skip=True)
        assert result == {}

    def test_mask_error_on_simple_fields(self):
        model = {
            "name": fields.String,
        }

        with pytest.raises(mask.MaskError):
            mask.apply(model, "name{notpossible}")

    def test_mask_error_on_list_field(self):
        model = {"nested": fields.List(fields.String)}

        with pytest.raises(mask.MaskError):
            mask.apply(model, "nested{notpossible}")


class MaskAPI(object):
    def test_marshal_with_honour_field_mask_header(self, app, client):
        api = Api(app)

        model = api.model(
            "Test",
            {
                "name": fields.String,
                "age": fields.Integer,
                "boolean": fields.Boolean,
            },
        )

        @api.route("/test/")
        class TestResource(Resource):
            @api.marshal_with(model)
            def get(self):
                return {"name": "John Doe", "age": 42, "boolean": True}

        data = client.get_json("/test/", headers={"X-Fields": "{name,age}"})
        assert data == {"name": "John Doe", "age": 42}

    def test_marshal_with_honour_field_mask_list(self, app, client):
        api = Api(app)

        model = api.model(
            "Test",
            {
                "name": fields.String,
                "age": fields.Integer,
                "boolean": fields.Boolean,
            },
        )

        @api.route("/test/")
        class TestResource(Resource):
            @api.marshal_with(model)
            def get(self):
                return [
                    {"name": "John Doe", "age": 42, "boolean": True},
                    {"name": "Jane Doe", "age": 33, "boolean": False},
                ]

        data = client.get_json("/test/", headers={"X-Fields": "{name,age}"})
        assert data == [
            {
                "name": "John Doe",
                "age": 42,
            },
            {
                "name": "Jane Doe",
                "age": 33,
            },
        ]

    def test_marshal_with_honour_complex_field_mask_header(self, app, client):
        api = Api(app)

        person = api.model("Person", person_fields)
        child = api.inherit("Child", person, {"attr": fields.String})

        family = api.model(
            "Family",
            {
                "father": fields.Nested(person),
                "mother": fields.Nested(person),
                "children": fields.List(fields.Nested(child)),
                "free": fields.List(fields.Raw),
            },
        )

        house = api.model(
            "House", {"family": fields.Nested(family, attribute="people")}
        )

        @api.route("/test/")
        class TestResource(Resource):
            @api.marshal_with(house)
            def get(self):
                return {
                    "people": {
                        "father": {"name": "John", "age": 42},
                        "mother": {"name": "Jane", "age": 42},
                        "children": [
                            {"name": "Jack", "age": 5, "attr": "value-1"},
                            {"name": "Julie", "age": 7, "attr": "value-2"},
                        ],
                        "free": [
                            {"key-1": "1-1", "key-2": "1-2"},
                            {"key-1": "2-1", "key-2": "2-2"},
                        ],
                    }
                }

        data = client.get_json(
            "/test/",
            headers={
                "X-Fields": "family{father{name},mother{age},children{name,attr},free{key-2}}"
            },
        )
        assert data == {
            "family": {
                "father": {"name": "John"},
                "mother": {"age": 42},
                "children": [
                    {"name": "Jack", "attr": "value-1"},
                    {"name": "Julie", "attr": "value-2"},
                ],
                "free": [{"key-2": "1-2"}, {"key-2": "2-2"}],
            }
        }

    def test_marshal_honour_field_mask(self, app):
        api = Api(app)

        model = api.model(
            "Test",
            {
                "name": fields.String,
                "age": fields.Integer,
                "boolean": fields.Boolean,
            },
        )

        data = {"name": "John Doe", "age": 42, "boolean": True}

        result = api.marshal(data, model, mask="{name,age}")

        assert result == {
            "name": "John Doe",
            "age": 42,
        }

    def test_marshal_with_honour_default_mask(self, app, client):
        api = Api(app)

        model = api.model(
            "Test",
            {
                "name": fields.String,
                "age": fields.Integer,
                "boolean": fields.Boolean,
            },
        )

        @api.route("/test/")
        class TestResource(Resource):
            @api.marshal_with(model, mask="{name,age}")
            def get(self):
                return {"name": "John Doe", "age": 42, "boolean": True}

        data = self.get_json("/test/")
        self.assertEqual(
            data,
            {
                "name": "John Doe",
                "age": 42,
            },
        )

    def test_marshal_with_honour_default_model_mask(self, app, client):
        api = Api(app)

        model = api.model(
            "Test",
            {
                "name": fields.String,
                "age": fields.Integer,
                "boolean": fields.Boolean,
            },
            mask="{name,age}",
        )

        @api.route("/test/")
        class TestResource(Resource):
            @api.marshal_with(model)
            def get(self):
                return {"name": "John Doe", "age": 42, "boolean": True}

        data = client.get_json("/test/")
        assert data == {"name": "John Doe", "age": 42}

    def test_marshal_with_honour_header_field_mask_with_default_model_mask(
        self, app, client
    ):
        api = Api(app)

        model = api.model(
            "Test",
            {
                "name": fields.String,
                "age": fields.Integer,
                "boolean": fields.Boolean,
            },
            mask="{name,age}",
        )

        @api.route("/test/")
        class TestResource(Resource):
            @api.marshal_with(model)
            def get(self):
                return {"name": "John Doe", "age": 42, "boolean": True}

        data = client.get_json("/test/", headers={"X-Fields": "{name}"})
        assert data == {"name": "John Doe"}

    def test_marshal_with_honour_header_default_mask_with_default_model_mask(
        self, app, client
    ):
        api = Api(app)

        model = api.model(
            "Test",
            {
                "name": fields.String,
                "age": fields.Integer,
                "boolean": fields.Boolean,
            },
            mask="{name,boolean}",
        )

        @api.route("/test/")
        class TestResource(Resource):
            @api.marshal_with(model, mask="{name}")
            def get(self):
                return {"name": "John Doe", "age": 42, "boolean": True}

        data = client.get_json("/test/")
        assert data == {"name": "John Doe"}

    def test_marshal_with_honour_header_field_mask_with_default_mask(self, app, client):
        api = Api(app)

        model = api.model(
            "Test",
            {
                "name": fields.String,
                "age": fields.Integer,
                "boolean": fields.Boolean,
            },
        )

        @api.route("/test/")
        class TestResource(Resource):
            @api.marshal_with(model, mask="{name,age}")
            def get(self):
                return {"name": "John Doe", "age": 42, "boolean": True}

        data = client.get_json("/test/", headers={"X-Fields": "{name}"})
        assert data == {"name": "John Doe"}

    def test_marshal_with_honour_header_field_mask_with_default_mask_and_default_model_mask(
        self, app, client
    ):
        api = Api(app)

        model = api.model(
            "Test",
            {
                "name": fields.String,
                "age": fields.Integer,
                "boolean": fields.Boolean,
            },
            mask="{name,boolean}",
        )

        @api.route("/test/")
        class TestResource(Resource):
            @api.marshal_with(model, mask="{name,age}")
            def get(self):
                return {"name": "John Doe", "age": 42, "boolean": True}

        data = client.get_json("/test/", headers={"X-Fields": "{name}"})
        assert data == {"name": "John Doe"}

    def test_marshal_with_honour_custom_field_mask(self, app, client):
        api = Api(app)

        model = api.model(
            "Test",
            {
                "name": fields.String,
                "age": fields.Integer,
                "boolean": fields.Boolean,
            },
        )

        @api.route("/test/")
        class TestResource(Resource):
            @api.marshal_with(model)
            def get(self):
                return {"name": "John Doe", "age": 42, "boolean": True}

        app.config["RESTX_MASK_HEADER"] = "X-Mask"
        data = client.get_json("/test/", headers={"X-Mask": "{name,age}"})

        assert data == {"name": "John Doe", "age": 42}

    def test_marshal_does_not_hit_unrequired_attributes(self, app, client):
        api = Api(app)

        model = api.model(
            "Person",
            {
                "name": fields.String,
                "age": fields.Integer,
                "boolean": fields.Boolean,
            },
        )

        class Person(object):
            def __init__(self, name, age):
                self.name = name
                self.age = age

            @property
            def boolean(self):
                raise Exception()

        @api.route("/test/")
        class TestResource(Resource):
            @api.marshal_with(model)
            def get(self):
                return Person("John Doe", 42)

        data = client.get_json("/test/", headers={"X-Fields": "{name,age}"})
        assert data == {"name": "John Doe", "age": 42}

    def test_marshal_with_skip_missing_fields(self, app, client):
        api = Api(app)

        model = api.model(
            "Test",
            {
                "name": fields.String,
                "age": fields.Integer,
            },
        )

        @api.route("/test/")
        class TestResource(Resource):
            @api.marshal_with(model)
            def get(self):
                return {
                    "name": "John Doe",
                    "age": 42,
                }

        data = client.get_json("/test/", headers={"X-Fields": "{name,missing}"})
        assert data == {"name": "John Doe"}

    def test_marshal_handle_inheritance(self, app):
        api = Api(app)

        person = api.model(
            "Person",
            {
                "name": fields.String,
                "age": fields.Integer,
            },
        )

        child = api.inherit(
            "Child",
            person,
            {
                "extra": fields.String,
            },
        )

        data = {"name": "John Doe", "age": 42, "extra": "extra"}

        values = (
            ("name", {"name": "John Doe"}),
            ("name,extra", {"name": "John Doe", "extra": "extra"}),
            ("extra", {"extra": "extra"}),
        )

        for value, expected in values:
            result = marshal(data, child, mask=value)
            assert result == expected

    def test_marshal_with_handle_polymorph(self, app, client):
        api = Api(app)

        parent = api.model(
            "Person",
            {
                "name": fields.String,
            },
        )

        child1 = api.inherit(
            "Child1",
            parent,
            {
                "extra1": fields.String,
            },
        )

        child2 = api.inherit(
            "Child2",
            parent,
            {
                "extra2": fields.String,
            },
        )

        class Child1(object):
            name = "child1"
            extra1 = "extra1"

        class Child2(object):
            name = "child2"
            extra2 = "extra2"

        mapping = {Child1: child1, Child2: child2}

        thing = api.model(
            "Thing",
            {
                "owner": fields.Polymorph(mapping),
            },
        )

        @api.route("/thing-1/")
        class Thing1Resource(Resource):
            @api.marshal_with(thing)
            def get(self):
                return {"owner": Child1()}

        @api.route("/thing-2/")
        class Thing2Resource(Resource):
            @api.marshal_with(thing)
            def get(self):
                return {"owner": Child2()}

        data = client.get_json("/thing-1/", headers={"X-Fields": "owner{name}"})
        assert data == {"owner": {"name": "child1"}}

        data = client.get_json("/thing-1/", headers={"X-Fields": "owner{extra1}"})
        assert data == {"owner": {"extra1": "extra1"}}

        data = client.get_json("/thing-2/", headers={"X-Fields": "owner{name}"})
        assert data == {"owner": {"name": "child2"}}

    def test_raise_400_on_invalid_mask(self, app, client):
        api = Api(app)

        model = api.model(
            "Test",
            {
                "name": fields.String,
                "age": fields.Integer,
            },
        )

        @api.route("/test/")
        class TestResource(Resource):
            @api.marshal_with(model)
            def get(self):
                pass

        response = client.get("/test/", headers={"X-Fields": "name{,missing}"})
        assert response.status_code == 400
        assert response.content_type == "application/json"


class SwaggerMaskHeaderTest(object):
    def test_marshal_with_expose_mask_header(self, app, client):
        api = Api(app)

        model = api.model(
            "Test",
            {
                "name": fields.String,
                "age": fields.Integer,
                "boolean": fields.Boolean,
            },
        )

        @api.route("/test/")
        class TestResource(Resource):
            @api.marshal_with(model)
            def get(self):
                return {"name": "John Doe", "age": 42, "boolean": True}

        specs = client.get_specs()
        op = specs["paths"]["/test/"]["get"]

        assert "parameters" in op
        assert len(op["parameters"]) == 1

        param = op["parameters"][0]

        assert param["name"] == "X-Fields"
        assert param["type"] == "string"
        assert param["format"] == "mask"
        assert param["in"] == "header"
        assert "required" not in param
        assert "default" not in param

    def test_marshal_with_expose_custom_mask_header(self, app, client):
        api = Api(app)

        model = api.model(
            "Test",
            {
                "name": fields.String,
                "age": fields.Integer,
                "boolean": fields.Boolean,
            },
        )

        @api.route("/test/")
        class TestResource(Resource):
            @api.marshal_with(model)
            def get(self):
                return {"name": "John Doe", "age": 42, "boolean": True}

        app.config["RESTX_MASK_HEADER"] = "X-Mask"
        specs = client.get_specs()

        op = specs["paths"]["/test/"]["get"]
        assert "parameters" in op
        assert len(op["parameters"]) == 1

        param = op["parameters"][0]
        assert param["name"] == "X-Mask"

    def test_marshal_with_disabling_mask_header(self, app, client):
        api = Api(app)

        model = api.model(
            "Test",
            {
                "name": fields.String,
                "age": fields.Integer,
                "boolean": fields.Boolean,
            },
        )

        @api.route("/test/")
        class TestResource(Resource):
            @api.marshal_with(model)
            def get(self):
                return {"name": "John Doe", "age": 42, "boolean": True}

        app.config["RESTX_MASK_SWAGGER"] = False
        specs = client.get_specs()

        op = specs["paths"]["/test/"]["get"]

        assert "parameters" not in op

    def test_is_only_exposed_on_marshal_with(self, app, client):
        api = Api(app)

        model = api.model(
            "Test",
            {
                "name": fields.String,
                "age": fields.Integer,
                "boolean": fields.Boolean,
            },
        )

        @api.route("/test/")
        class TestResource(Resource):
            def get(self):
                return api.marshal(
                    {"name": "John Doe", "age": 42, "boolean": True}, model
                )

        specs = client.get_specs()
        op = specs["paths"]["/test/"]["get"]

        assert "parameters" not in op

    def test_marshal_with_expose_default_mask_header(self, app, client):
        api = Api(app)

        model = api.model(
            "Test",
            {
                "name": fields.String,
                "age": fields.Integer,
                "boolean": fields.Boolean,
            },
        )

        @api.route("/test/")
        class TestResource(Resource):
            @api.marshal_with(model, mask="{name,age}")
            def get(self):
                pass

        specs = client.get_specs()
        op = specs["paths"]["/test/"]["get"]

        assert "parameters" in op
        assert len(op["parameters"]) == 1

        param = op["parameters"][0]

        assert param["name"] == "X-Fields"
        assert param["type"] == "string"
        assert param["format"] == "mask"
        assert param["default"] == "{name,age}"
        assert param["in"] == "header"
        assert "required" not in param

    def test_marshal_with_expose_default_model_mask_header(self, app, client):
        api = Api(app)

        model = api.model(
            "Test",
            {
                "name": fields.String,
                "age": fields.Integer,
                "boolean": fields.Boolean,
            },
            mask="{name,age}",
        )

        @api.route("/test/")
        class TestResource(Resource):
            @api.marshal_with(model)
            def get(self):
                pass

        specs = client.get_specs()
        definition = specs["definitions"]["Test"]
        assert "x-mask" in definition
        assert definition["x-mask"] == "{name,age}"



import copy
import pytest

from collections import OrderedDict

from flask_restx import fields, Model, OrderedModel, SchemaModel


class ModelTest(object):
    def test_model_as_flat_dict(self):
        model = Model(
            "Person",
            {
                "name": fields.String,
                "age": fields.Integer,
                "birthdate": fields.DateTime,
            },
        )

        assert isinstance(model, dict)
        assert not isinstance(model, OrderedDict)

        assert model.__schema__ == {
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "birthdate": {"type": "string", "format": "date-time"},
            },
            "type": "object",
        }

    def test_model_as_ordered_dict(self):
        model = OrderedModel(
            "Person",
            [
                ("name", fields.String),
                ("age", fields.Integer),
                ("birthdate", fields.DateTime),
            ],
        )

        assert isinstance(model, OrderedDict)

        assert model.__schema__ == {
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "birthdate": {"type": "string", "format": "date-time"},
            },
        }

    def test_model_as_nested_dict(self):
        address = Model(
            "Address",
            {
                "road": fields.String,
            },
        )

        person = Model(
            "Person",
            {
                "name": fields.String,
                "age": fields.Integer,
                "birthdate": fields.DateTime,
                "address": fields.Nested(address),
            },
        )

        assert person.__schema__ == {
            # 'required': ['address'],
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "birthdate": {"type": "string", "format": "date-time"},
                "address": {
                    "$ref": "#/definitions/Address",
                },
            },
            "type": "object",
        }

        assert address.__schema__ == {
            "properties": {
                "road": {"type": "string"},
            },
            "type": "object",
        }

    def test_model_as_dict_with_list(self):
        model = Model(
            "Person",
            {
                "name": fields.String,
                "age": fields.Integer,
                "tags": fields.List(fields.String),
            },
        )

        assert model.__schema__ == {
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "tags": {"type": "array", "items": {"type": "string"}},
            },
            "type": "object",
        }

    def test_model_as_nested_dict_with_list(self):
        address = Model(
            "Address",
            {
                "road": fields.String,
            },
        )

        person = Model(
            "Person",
            {
                "name": fields.String,
                "age": fields.Integer,
                "birthdate": fields.DateTime,
                "addresses": fields.List(fields.Nested(address)),
            },
        )

        assert person.__schema__ == {
            # 'required': ['address'],
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "birthdate": {"type": "string", "format": "date-time"},
                "addresses": {
                    "type": "array",
                    "items": {
                        "$ref": "#/definitions/Address",
                    },
                },
            },
            "type": "object",
        }

        assert address.__schema__ == {
            "properties": {
                "road": {"type": "string"},
            },
            "type": "object",
        }

    def test_model_with_required(self):
        model = Model(
            "Person",
            {
                "name": fields.String(required=True),
                "age": fields.Integer,
                "birthdate": fields.DateTime(required=True),
            },
        )

        assert model.__schema__ == {
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "birthdate": {"type": "string", "format": "date-time"},
            },
            "required": ["birthdate", "name"],
            "type": "object",
        }

    def test_model_as_nested_dict_and_required(self):
        address = Model(
            "Address",
            {
                "road": fields.String,
            },
        )

        person = Model(
            "Person",
            {
                "name": fields.String,
                "age": fields.Integer,
                "birthdate": fields.DateTime,
                "address": fields.Nested(address, required=True),
            },
        )

        assert person.__schema__ == {
            "required": ["address"],
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "birthdate": {"type": "string", "format": "date-time"},
                "address": {
                    "$ref": "#/definitions/Address",
                },
            },
            "type": "object",
        }

        assert address.__schema__ == {
            "properties": {
                "road": {"type": "string"},
            },
            "type": "object",
        }

    def test_model_with_discriminator(self):
        model = Model(
            "Person",
            {
                "name": fields.String(discriminator=True),
                "age": fields.Integer,
            },
        )

        assert model.__schema__ == {
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
            },
            "discriminator": "name",
            "required": ["name"],
            "type": "object",
        }

    def test_model_with_discriminator_override_require(self):
        model = Model(
            "Person",
            {
                "name": fields.String(discriminator=True, required=False),
                "age": fields.Integer,
            },
        )

        assert model.__schema__ == {
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
            },
            "discriminator": "name",
            "required": ["name"],
            "type": "object",
        }

    def test_model_deepcopy(self):
        parent = Model(
            "Person",
            {
                "name": fields.String,
                "age": fields.Integer(description="foo"),
            },
        )

        child = parent.inherit(
            "Child",
            {
                "extra": fields.String,
            },
        )

        parent_copy = copy.deepcopy(parent)

        assert parent_copy["age"].description == "foo"

        parent_copy["age"].description = "bar"

        assert parent["age"].description == "foo"
        assert parent_copy["age"].description == "bar"

        child = parent.inherit(
            "Child",
            {
                "extra": fields.String,
            },
        )

        child_copy = copy.deepcopy(child)
        assert child_copy.__parents__[0] == parent

    def test_clone_from_instance(self):
        parent = Model(
            "Parent",
            {
                "name": fields.String,
                "age": fields.Integer,
                "birthdate": fields.DateTime,
            },
        )

        child = parent.clone(
            "Child",
            {
                "extra": fields.String,
            },
        )

        assert child.__schema__ == {
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "birthdate": {"type": "string", "format": "date-time"},
                "extra": {"type": "string"},
            },
            "type": "object",
        }

    def test_clone_from_class(self):
        parent = Model(
            "Parent",
            {
                "name": fields.String,
                "age": fields.Integer,
                "birthdate": fields.DateTime,
            },
        )

        child = Model.clone(
            "Child",
            parent,
            {
                "extra": fields.String,
            },
        )

        assert child.__schema__ == {
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "birthdate": {"type": "string", "format": "date-time"},
                "extra": {"type": "string"},
            },
            "type": "object",
        }

    def test_clone_from_instance_with_multiple_parents(self):
        grand_parent = Model(
            "GrandParent",
            {
                "grand_parent": fields.String,
            },
        )

        parent = Model(
            "Parent",
            {
                "name": fields.String,
                "age": fields.Integer,
                "birthdate": fields.DateTime,
            },
        )

        child = grand_parent.clone(
            "Child",
            parent,
            {
                "extra": fields.String,
            },
        )

        assert child.__schema__ == {
            "properties": {
                "grand_parent": {"type": "string"},
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "birthdate": {"type": "string", "format": "date-time"},
                "extra": {"type": "string"},
            },
            "type": "object",
        }

    def test_clone_from_class_with_multiple_parents(self):
        grand_parent = Model(
            "GrandParent",
            {
                "grand_parent": fields.String,
            },
        )

        parent = Model(
            "Parent",
            {
                "name": fields.String,
                "age": fields.Integer,
                "birthdate": fields.DateTime,
            },
        )

        child = Model.clone(
            "Child",
            grand_parent,
            parent,
            {
                "extra": fields.String,
            },
        )

        assert child.__schema__ == {
            "properties": {
                "grand_parent": {"type": "string"},
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "birthdate": {"type": "string", "format": "date-time"},
                "extra": {"type": "string"},
            },
            "type": "object",
        }

    def test_inherit_from_instance(self):
        parent = Model(
            "Parent",
            {
                "name": fields.String,
                "age": fields.Integer,
            },
        )

        child = parent.inherit(
            "Child",
            {
                "extra": fields.String,
            },
        )

        assert parent.__schema__ == {
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
            },
            "type": "object",
        }
        assert child.__schema__ == {
            "allOf": [
                {"$ref": "#/definitions/Parent"},
                {"properties": {"extra": {"type": "string"}}, "type": "object"},
            ]
        }

    def test_inherit_from_class(self):
        parent = Model(
            "Parent",
            {
                "name": fields.String,
                "age": fields.Integer,
            },
        )

        child = Model.inherit(
            "Child",
            parent,
            {
                "extra": fields.String,
            },
        )

        assert parent.__schema__ == {
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
            },
            "type": "object",
        }
        assert child.__schema__ == {
            "allOf": [
                {"$ref": "#/definitions/Parent"},
                {"properties": {"extra": {"type": "string"}}, "type": "object"},
            ]
        }

    def test_inherit_from_class_from_multiple_parents(self):
        grand_parent = Model(
            "GrandParent",
            {
                "grand_parent": fields.String,
            },
        )

        parent = Model(
            "Parent",
            {
                "name": fields.String,
                "age": fields.Integer,
            },
        )

        child = Model.inherit(
            "Child",
            grand_parent,
            parent,
            {
                "extra": fields.String,
            },
        )

        assert child.__schema__ == {
            "allOf": [
                {"$ref": "#/definitions/GrandParent"},
                {"$ref": "#/definitions/Parent"},
                {"properties": {"extra": {"type": "string"}}, "type": "object"},
            ]
        }

    def test_inherit_from_instance_from_multiple_parents(self):
        grand_parent = Model(
            "GrandParent",
            {
                "grand_parent": fields.String,
            },
        )

        parent = Model(
            "Parent",
            {
                "name": fields.String,
                "age": fields.Integer,
            },
        )

        child = grand_parent.inherit(
            "Child",
            parent,
            {
                "extra": fields.String,
            },
        )

        assert child.__schema__ == {
            "allOf": [
                {"$ref": "#/definitions/GrandParent"},
                {"$ref": "#/definitions/Parent"},
                {"properties": {"extra": {"type": "string"}}, "type": "object"},
            ]
        }

    # def test_inherit_inline(self):
    #     parent = Model('Person', {
    #         'name': fields.String,
    #         'age': fields.Integer,
    #     })
    #
    #     child = self.api.inherit('Child', parent, {
    #         'extra': fields.String,
    #     })
    #
    #     Model('Output', {
    #         'child': fields.Nested(child),
    #         'children': fields.List(fields.Nested(child))
    #     })
    #
    #     self.assertIn('Person', Models)
    #     self.assertIn('Child', Models)

    def test_polymorph_inherit_common_ancestor(self):
        class Child1:
            pass

        class Child2:
            pass

        parent = Model(
            "Person",
            {
                "name": fields.String,
                "age": fields.Integer,
            },
        )

        child1 = parent.inherit(
            "Child1",
            {
                "extra1": fields.String,
            },
        )

        child2 = parent.inherit(
            "Child2",
            {
                "extra2": fields.String,
            },
        )

        mapping = {
            Child1: child1,
            Child2: child2,
        }

        output = Model("Output", {"child": fields.Polymorph(mapping)})

        # Should use the common ancestor
        assert output.__schema__ == {
            "properties": {
                "child": {"$ref": "#/definitions/Person"},
            },
            "type": "object",
        }

    def test_validate(self):
        from jsonschema import FormatChecker
        from werkzeug.exceptions import BadRequest

        class IPAddress(fields.Raw):
            __schema_type__ = "string"
            __schema_format__ = "ipv4"

        data = {"ip": "192.168.1"}
        model = Model("MyModel", {"ip": IPAddress()})

        # Test that validate without a FormatChecker does not check if a
        # primitive type conforms to the defined format property
        assert model.validate(data) is None

        # Test that validate with a FormatChecker enforces the check of the
        # format property and throws an error if invalid
        with pytest.raises(BadRequest):
            model.validate(data, format_checker=FormatChecker())


class ModelSchemaTestCase(object):
    def test_model_schema(self):
        address = SchemaModel(
            "Address",
            {
                "properties": {
                    "road": {"type": "string"},
                },
                "type": "object",
            },
        )

        person = SchemaModel(
            "Person",
            {
                # 'required': ['address'],
                "properties": {
                    "name": {"type": "string"},
                    "age": {"type": "integer"},
                    "birthdate": {"type": "string", "format": "date-time"},
                    "address": {
                        "$ref": "#/definitions/Address",
                    },
                },
                "type": "object",
            },
        )

        assert person.__schema__ == {
            # 'required': ['address'],
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "birthdate": {"type": "string", "format": "date-time"},
                "address": {
                    "$ref": "#/definitions/Address",
                },
            },
            "type": "object",
        }

        assert address.__schema__ == {
            "properties": {
                "road": {"type": "string"},
            },
            "type": "object",
        }


class ModelDeprecattionsTest(object):
    def test_extend_is_deprecated(self):
        parent = Model(
            "Parent",
            {
                "name": fields.String,
                "age": fields.Integer,
                "birthdate": fields.DateTime,
            },
        )

        with pytest.warns(DeprecationWarning):
            child = parent.extend(
                "Child",
                {
                    "extra": fields.String,
                },
            )

        assert child.__schema__ == {
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "birthdate": {"type": "string", "format": "date-time"},
                "extra": {"type": "string"},
            },
            "type": "object",
        }



import pytest

from flask_restx import marshal, marshal_with, marshal_with_field, fields, Api, Resource

from collections import OrderedDict


# Add a dummy Resource to verify that the app is properly set.
class HelloWorld(Resource):
    def get(self):
        return {}


class MarshallingTest(object):
    def test_marshal(self):
        model = OrderedDict([("foo", fields.Raw)])
        marshal_dict = OrderedDict([("foo", "bar"), ("bat", "baz")])
        output = marshal(marshal_dict, model)
        assert isinstance(output, dict)
        assert not isinstance(output, OrderedDict)
        assert output == {"foo": "bar"}

    def test_marshal_wildcard_nested(self):
        nest = fields.Nested(
            OrderedDict([("thumbnail", fields.String), ("video", fields.String)])
        )
        wild = fields.Wildcard(nest)
        wildcard_fields = OrderedDict([("*", wild)])
        model = OrderedDict([("preview", fields.Nested(wildcard_fields))])
        sub_dict = OrderedDict(
            [
                ("9:16", {"thumbnail": 24, "video": 12}),
                ("16:9", {"thumbnail": 25, "video": 11}),
                ("1:1", {"thumbnail": 26, "video": 10}),
            ]
        )
        marshal_dict = OrderedDict([("preview", sub_dict)])
        output = marshal(marshal_dict, model)
        assert output == {
            "preview": {
                "1:1": {"thumbnail": "26", "video": "10"},
                "16:9": {"thumbnail": "25", "video": "11"},
                "9:16": {"thumbnail": "24", "video": "12"},
            }
        }

    def test_marshal_wildcard_list(self):
        wild = fields.Wildcard(fields.List(fields.String))
        wildcard_fields = OrderedDict([("*", wild)])
        model = OrderedDict([("preview", fields.Nested(wildcard_fields))])
        sub_dict = OrderedDict(
            [("1:1", [1, 2, 3]), ("16:9", [4, 5, 6]), ("9:16", [7, 8, 9])]
        )
        marshal_dict = OrderedDict([("preview", sub_dict)])
        output = marshal(marshal_dict, model)
        assert output == {
            "preview": {
                "9:16": ["7", "8", "9"],
                "16:9": ["4", "5", "6"],
                "1:1": ["1", "2", "3"],
            }
        }

    def test_marshal_with_envelope(self):
        model = OrderedDict([("foo", fields.Raw)])
        marshal_dict = OrderedDict([("foo", "bar"), ("bat", "baz")])
        output = marshal(marshal_dict, model, envelope="hey")
        assert output == {"hey": {"foo": "bar"}}

    def test_marshal_wildcard_with_envelope(self):
        wild = fields.Wildcard(fields.String)
        model = OrderedDict([("foo", fields.Raw), ("*", wild)])
        marshal_dict = OrderedDict(
            [("foo", {"bat": "baz"}), ("a", "toto"), ("b", "tata")]
        )
        output = marshal(marshal_dict, model, envelope="hey")
        assert output == {"hey": {"a": "toto", "b": "tata", "foo": {"bat": "baz"}}}

    def test_marshal_with_skip_none(self):
        model = OrderedDict(
            [("foo", fields.Raw), ("bat", fields.Raw), ("qux", fields.Raw)]
        )
        marshal_dict = OrderedDict([("foo", "bar"), ("bat", None)])
        output = marshal(marshal_dict, model, skip_none=True)
        assert output == {"foo": "bar"}

    def test_marshal_wildcard_with_skip_none(self):
        wild = fields.Wildcard(fields.String)
        model = OrderedDict([("foo", fields.Raw), ("*", wild)])
        marshal_dict = OrderedDict(
            [("foo", None), ("bat", None), ("baz", "biz"), ("bar", None)]
        )
        output = marshal(marshal_dict, model, skip_none=True)
        assert output == {"baz": "biz"}

    def test_marshal_decorator(self):
        model = OrderedDict([("foo", fields.Raw)])

        @marshal_with(model)
        def try_me():
            return OrderedDict([("foo", "bar"), ("bat", "baz")])

        assert try_me() == {"foo": "bar"}

    def test_marshal_decorator_with_envelope(self):
        model = OrderedDict([("foo", fields.Raw)])

        @marshal_with(model, envelope="hey")
        def try_me():
            return OrderedDict([("foo", "bar"), ("bat", "baz")])

        assert try_me() == {"hey": {"foo": "bar"}}

    def test_marshal_decorator_with_skip_none(self):
        model = OrderedDict(
            [("foo", fields.Raw), ("bat", fields.Raw), ("qux", fields.Raw)]
        )

        @marshal_with(model, skip_none=True)
        def try_me():
            return OrderedDict([("foo", "bar"), ("bat", None)])

        assert try_me() == {"foo": "bar"}

    def test_marshal_decorator_tuple(self):
        model = OrderedDict([("foo", fields.Raw)])

        @marshal_with(model)
        def try_me():
            headers = {"X-test": 123}
            return OrderedDict([("foo", "bar"), ("bat", "baz")]), 200, headers

        assert try_me() == ({"foo": "bar"}, 200, {"X-test": 123})

    def test_marshal_decorator_tuple_with_envelope(self):
        model = OrderedDict([("foo", fields.Raw)])

        @marshal_with(model, envelope="hey")
        def try_me():
            headers = {"X-test": 123}
            return OrderedDict([("foo", "bar"), ("bat", "baz")]), 200, headers

        assert try_me() == ({"hey": {"foo": "bar"}}, 200, {"X-test": 123})

    def test_marshal_decorator_tuple_with_skip_none(self):
        model = OrderedDict(
            [("foo", fields.Raw), ("bat", fields.Raw), ("qux", fields.Raw)]
        )

        @marshal_with(model, skip_none=True)
        def try_me():
            headers = {"X-test": 123}
            return OrderedDict([("foo", "bar"), ("bat", None)]), 200, headers

        assert try_me() == ({"foo": "bar"}, 200, {"X-test": 123})

    def test_marshal_field_decorator(self):
        model = fields.Raw

        @marshal_with_field(model)
        def try_me():
            return "foo"

        assert try_me() == "foo"

    def test_marshal_field_decorator_tuple(self):
        model = fields.Raw

        @marshal_with_field(model)
        def try_me():
            return "foo", 200, {"X-test": 123}

        assert try_me() == ("foo", 200, {"X-test": 123})

    def test_marshal_field(self):
        model = OrderedDict({"foo": fields.Raw()})
        marshal_fields = OrderedDict([("foo", "bar"), ("bat", "baz")])
        output = marshal(marshal_fields, model)
        assert output == {"foo": "bar"}

    def test_marshal_tuple(self):
        model = OrderedDict({"foo": fields.Raw})
        marshal_fields = OrderedDict([("foo", "bar"), ("bat", "baz")])
        output = marshal((marshal_fields,), model)
        assert output == [{"foo": "bar"}]

    def test_marshal_tuple_with_envelope(self):
        model = OrderedDict({"foo": fields.Raw})
        marshal_fields = OrderedDict([("foo", "bar"), ("bat", "baz")])
        output = marshal((marshal_fields,), model, envelope="hey")
        assert output == {"hey": [{"foo": "bar"}]}

    def test_marshal_tuple_with_skip_none(self):
        model = OrderedDict(
            [("foo", fields.Raw), ("bat", fields.Raw), ("qux", fields.Raw)]
        )
        marshal_fields = OrderedDict([("foo", "bar"), ("bat", None)])
        output = marshal((marshal_fields,), model, skip_none=True)
        assert output == [{"foo": "bar"}]

    def test_marshal_nested(self):
        model = {
            "foo": fields.Raw,
            "fee": fields.Nested({"fye": fields.String}),
        }

        marshal_fields = {
            "foo": "bar",
            "bat": "baz",
            "fee": {"fye": "fum"},
        }
        expected = {
            "foo": "bar",
            "fee": {"fye": "fum"},
        }

        output = marshal(marshal_fields, model)

        assert output == expected

    def test_marshal_ordered(self):
        model = OrderedDict(
            [("foo", fields.Raw), ("baz", fields.Raw), ("bar", fields.Raw)]
        )
        marshal_fields = {"foo": 1, "baz": 2, "bar": 3}
        expected_ordered = OrderedDict([("foo", 1), ("baz", 2), ("bar", 3)])
        ordered_output = marshal(marshal_fields, model, ordered=True)
        assert ordered_output == expected_ordered
        unordered_output = marshal(marshal_fields, model)
        assert not isinstance(unordered_output, OrderedDict)

    def test_marshal_nested_ordered(self):
        model = OrderedDict(
            [
                ("foo", fields.Raw),
                (
                    "fee",
                    fields.Nested(
                        {
                            "fye": fields.String,
                        }
                    ),
                ),
            ]
        )

        marshal_fields = {
            "foo": "bar",
            "bat": "baz",
            "fee": {"fye": "fum"},
        }
        expected = OrderedDict([("foo", "bar"), ("fee", OrderedDict([("fye", "fum")]))])

        output = marshal(marshal_fields, model, ordered=True)

        assert isinstance(output, OrderedDict)
        assert output == expected
        assert isinstance(output["fee"], OrderedDict)

    def test_marshal_nested_with_non_null(self):
        model = OrderedDict(
            [
                ("foo", fields.Raw),
                (
                    "fee",
                    fields.Nested(
                        OrderedDict([("fye", fields.String), ("blah", fields.String)]),
                        allow_null=False,
                    ),
                ),
            ]
        )
        marshal_fields = [OrderedDict([("foo", "bar"), ("bat", "baz"), ("fee", None)])]
        output = marshal(marshal_fields, model)
        expected = [
            OrderedDict(
                [("foo", "bar"), ("fee", OrderedDict([("fye", None), ("blah", None)]))]
            )
        ]
        assert output == expected

    def test_marshal_nested_with_null(self):
        model = OrderedDict(
            [
                ("foo", fields.Raw),
                (
                    "fee",
                    fields.Nested(
                        OrderedDict([("fye", fields.String), ("blah", fields.String)]),
                        allow_null=True,
                    ),
                ),
            ]
        )
        marshal_fields = OrderedDict([("foo", "bar"), ("bat", "baz"), ("fee", None)])
        output = marshal(marshal_fields, model)
        expected = OrderedDict([("foo", "bar"), ("fee", None)])
        assert output == expected

    def test_marshal_nested_with_skip_none(self):
        model = OrderedDict(
            [
                ("foo", fields.Raw),
                (
                    "fee",
                    fields.Nested(
                        OrderedDict([("fye", fields.String)]), skip_none=True
                    ),
                ),
            ]
        )
        marshal_fields = OrderedDict([("foo", "bar"), ("bat", "baz"), ("fee", None)])
        output = marshal(marshal_fields, model, skip_none=True)
        expected = OrderedDict([("foo", "bar")])
        assert output == expected

    def test_allow_null_presents_data(self):
        model = OrderedDict(
            [
                ("foo", fields.Raw),
                (
                    "fee",
                    fields.Nested(
                        OrderedDict([("fye", fields.String), ("blah", fields.String)]),
                        allow_null=True,
                    ),
                ),
            ]
        )
        marshal_fields = OrderedDict(
            [("foo", "bar"), ("bat", "baz"), ("fee", {"blah": "cool"})]
        )
        output = marshal(marshal_fields, model)
        expected = OrderedDict(
            [("foo", "bar"), ("fee", OrderedDict([("fye", None), ("blah", "cool")]))]
        )
        assert output == expected

    def test_skip_none_presents_data(self):
        model = OrderedDict(
            [
                ("foo", fields.Raw),
                (
                    "fee",
                    fields.Nested(
                        OrderedDict(
                            [
                                ("fye", fields.String),
                                ("blah", fields.String),
                                ("foe", fields.String),
                            ]
                        ),
                        skip_none=True,
                    ),
                ),
            ]
        )
        marshal_fields = OrderedDict(
            [("foo", "bar"), ("bat", "baz"), ("fee", {"blah": "cool", "foe": None})]
        )
        output = marshal(marshal_fields, model)
        expected = OrderedDict(
            [("foo", "bar"), ("fee", OrderedDict([("blah", "cool")]))]
        )
        assert output == expected

    def test_marshal_nested_property(self):
        class TestObject(object):
            @property
            def fee(self):
                return {"blah": "cool"}

        model = OrderedDict(
            [
                ("foo", fields.Raw),
                (
                    "fee",
                    fields.Nested(
                        OrderedDict([("fye", fields.String), ("blah", fields.String)]),
                        allow_null=True,
                    ),
                ),
            ]
        )
        obj = TestObject()
        obj.foo = "bar"
        obj.bat = "baz"
        output = marshal([obj], model)
        expected = [
            OrderedDict(
                [
                    ("foo", "bar"),
                    ("fee", OrderedDict([("fye", None), ("blah", "cool")])),
                ]
            )
        ]
        assert output == expected

    def test_marshal_nested_property_with_skip_none(self):
        class TestObject(object):
            @property
            def fee(self):
                return {"blah": "cool", "foe": None}

        model = OrderedDict(
            [
                ("foo", fields.Raw),
                (
                    "fee",
                    fields.Nested(
                        OrderedDict(
                            [
                                ("fye", fields.String),
                                ("blah", fields.String),
                                ("foe", fields.String),
                            ]
                        ),
                        skip_none=True,
                    ),
                ),
            ]
        )
        obj = TestObject()
        obj.foo = "bar"
        obj.bat = "baz"
        output = marshal([obj], model)
        expected = [
            OrderedDict([("foo", "bar"), ("fee", OrderedDict([("blah", "cool")]))])
        ]
        assert output == expected

    def test_marshal_list(self):
        model = OrderedDict([("foo", fields.Raw), ("fee", fields.List(fields.String))])
        marshal_fields = OrderedDict(
            [("foo", "bar"), ("bat", "baz"), ("fee", ["fye", "fum"])]
        )
        output = marshal(marshal_fields, model)
        expected = OrderedDict([("foo", "bar"), ("fee", (["fye", "fum"]))])
        assert output == expected

    def test_marshal_list_of_nesteds(self):
        model = OrderedDict(
            [
                ("foo", fields.Raw),
                ("fee", fields.List(fields.Nested({"fye": fields.String}))),
            ]
        )
        marshal_fields = OrderedDict(
            [("foo", "bar"), ("bat", "baz"), ("fee", {"fye": "fum"})]
        )
        output = marshal(marshal_fields, model)
        expected = OrderedDict(
            [("foo", "bar"), ("fee", [OrderedDict([("fye", "fum")])])]
        )
        assert output == expected

    def test_marshal_list_of_lists(self):
        model = OrderedDict(
            [("foo", fields.Raw), ("fee", fields.List(fields.List(fields.String)))]
        )
        marshal_fields = OrderedDict(
            [("foo", "bar"), ("bat", "baz"), ("fee", [["fye"], ["fum"]])]
        )
        output = marshal(marshal_fields, model)
        expected = OrderedDict([("foo", "bar"), ("fee", [["fye"], ["fum"]])])
        assert output == expected

    def test_marshal_nested_dict(self):
        model = OrderedDict(
            [
                ("foo", fields.Raw),
                (
                    "bar",
                    OrderedDict(
                        [
                            ("a", fields.Raw),
                            ("b", fields.Raw),
                        ]
                    ),
                ),
            ]
        )
        marshal_fields = OrderedDict(
            [
                ("foo", "foo-val"),
                ("bar", "bar-val"),
                ("bat", "bat-val"),
                ("a", 1),
                ("b", 2),
                ("c", 3),
            ]
        )
        output = marshal(marshal_fields, model)
        expected = OrderedDict(
            [("foo", "foo-val"), ("bar", OrderedDict([("a", 1), ("b", 2)]))]
        )
        assert output == expected

    @pytest.mark.options(debug=True)
    def test_will_prettyprint_json_in_debug_mode(self, app, client):
        api = Api(app)

        class Foo1(Resource):
            def get(self):
                return {"foo": "bar", "baz": "asdf"}

        api.add_resource(Foo1, "/foo", endpoint="bar")

        foo = client.get("/foo")

        # Python's dictionaries have random order (as of "new" Pythons,
        # anyway), so we can't verify the actual output here.  We just
        # assert that they're properly prettyprinted.
        lines = foo.data.splitlines()
        lines = [line.decode() for line in lines]
        assert "{" == lines[0]
        assert lines[1].startswith("    ")
        assert lines[2].startswith("    ")
        assert "}" == lines[3]

        # Assert our trailing newline.
        assert foo.data.endswith(b"\n")

    def test_json_float_marshalled(self, app, client):
        api = Api(app)

        class FooResource(Resource):
            fields = {"foo": fields.Float}

            def get(self):
                return marshal({"foo": 3.0}, self.fields)

        api.add_resource(FooResource, "/api")

        resp = client.get("/api")
        assert resp.status_code == 200
        assert resp.data.decode("utf-8") == '{"foo": 3.0}\n'



import pytest

from flask import url_for, Blueprint
from werkzeug.routing import BuildError

import flask_restx as restx


class APIDocTest(object):
    def test_default_apidoc_on_root(self, app, client):
        restx.Api(app, version="1.0")

        assert url_for("doc") == url_for("root")

        response = client.get(url_for("doc"))
        assert response.status_code == 200
        assert response.content_type == "text/html; charset=utf-8"

    def test_default_apidoc_on_root_lazy(self, app, client):
        api = restx.Api(version="1.0")
        api.init_app(app)

        assert url_for("doc") == url_for("root")

        response = client.get(url_for("doc"))
        assert response.status_code == 200
        assert response.content_type == "text/html; charset=utf-8"

    def test_default_apidoc_on_root_with_blueprint(self, app, client):
        blueprint = Blueprint("api", __name__, url_prefix="/api")
        restx.Api(blueprint, version="1.0")
        app.register_blueprint(blueprint)

        assert url_for("api.doc") == url_for("api.root")

        response = client.get(url_for("api.doc"))
        assert response.status_code == 200
        assert response.content_type == "text/html; charset=utf-8"

    def test_apidoc_with_custom_validator(self, app, client):
        app.config["SWAGGER_VALIDATOR_URL"] = "http://somewhere.com/validator"
        restx.Api(app, version="1.0")

        response = client.get(url_for("doc"))
        assert response.status_code == 200
        assert response.content_type == "text/html; charset=utf-8"
        assert 'validatorUrl: "http://somewhere.com/validator" || null,' in str(
            response.data
        )

    def test_apidoc_doc_expansion_parameter(self, app, client):
        restx.Api(app)

        response = client.get(url_for("doc"))
        assert 'docExpansion: "none"' in str(response.data)

        app.config["SWAGGER_UI_DOC_EXPANSION"] = "list"
        response = client.get(url_for("doc"))
        assert 'docExpansion: "list"' in str(response.data)

        app.config["SWAGGER_UI_DOC_EXPANSION"] = "full"
        response = client.get(url_for("doc"))
        assert 'docExpansion: "full"' in str(response.data)

    def test_apidoc_doc_display_operation_id(self, app, client):
        restx.Api(app)

        response = client.get(url_for("doc"))
        assert "displayOperationId: false" in str(response.data)

        app.config["SWAGGER_UI_OPERATION_ID"] = False
        response = client.get(url_for("doc"))
        assert "displayOperationId: false" in str(response.data)

        app.config["SWAGGER_UI_OPERATION_ID"] = True
        response = client.get(url_for("doc"))
        assert "displayOperationId: true" in str(response.data)

    def test_apidoc_doc_display_request_duration(self, app, client):
        restx.Api(app)

        response = client.get(url_for("doc"))
        assert "displayRequestDuration: false" in str(response.data)

        app.config["SWAGGER_UI_REQUEST_DURATION"] = False
        response = client.get(url_for("doc"))
        assert "displayRequestDuration: false" in str(response.data)

        app.config["SWAGGER_UI_REQUEST_DURATION"] = True
        response = client.get(url_for("doc"))
        assert "displayRequestDuration: true" in str(response.data)

    def test_custom_apidoc_url(self, app, client):
        restx.Api(app, version="1.0", doc="/doc/")

        doc_url = url_for("doc")
        root_url = url_for("root")

        assert doc_url != root_url

        response = client.get(root_url)
        assert response.status_code == 404

        assert doc_url == "/doc/"
        response = client.get(doc_url)
        assert response.status_code == 200
        assert response.content_type == "text/html; charset=utf-8"

    def test_custom_api_prefix(self, app, client):
        prefix = "/api"
        api = restx.Api(app, prefix=prefix)
        api.namespace("resource")
        assert url_for("root") == prefix

    def test_custom_apidoc_page(self, app, client):
        api = restx.Api(app, version="1.0")
        content = "My Custom API Doc"

        @api.documentation
        def api_doc():
            return content

        response = client.get(url_for("doc"))
        assert response.status_code == 200
        assert response.data.decode("utf8") == content

    def test_custom_apidoc_page_lazy(self, app, client):
        blueprint = Blueprint("api", __name__, url_prefix="/api")
        api = restx.Api(blueprint, version="1.0")
        content = "My Custom API Doc"

        @api.documentation
        def api_doc():
            return content

        app.register_blueprint(blueprint)

        response = client.get(url_for("api.doc"))
        assert response.status_code == 200
        assert response.data.decode("utf8") == content

    def test_disabled_apidoc(self, app, client):
        restx.Api(app, version="1.0", doc=False)

        with pytest.raises(BuildError):
            url_for("doc")

        response = client.get(url_for("root"))
        assert response.status_code == 404



# -*- coding: utf-8 -*-
import re

import flask_restx as restx

from flask_restx import Namespace, Model, OrderedModel


class NamespaceTest(object):
    def test_parser(self):
        api = Namespace("test")
        assert isinstance(api.parser(), restx.reqparse.RequestParser)

    def test_doc_decorator(self):
        api = Namespace("test")
        params = {"q": {"description": "some description"}}

        @api.doc(params=params)
        class TestResource(restx.Resource):
            pass

        assert hasattr(TestResource, "__apidoc__")
        assert TestResource.__apidoc__ == {"params": params}

    def test_doc_with_inheritance(self):
        api = Namespace("test")
        base_params = {
            "q": {
                "description": "some description",
                "type": "string",
                "paramType": "query",
            }
        }
        child_params = {
            "q": {"description": "some new description"},
            "other": {"description": "another param"},
        }

        @api.doc(params=base_params)
        class BaseResource(restx.Resource):
            pass

        @api.doc(params=child_params)
        class TestResource(BaseResource):
            pass

        assert TestResource.__apidoc__ == {
            "params": {
                "q": {
                    "description": "some new description",
                    "type": "string",
                    "paramType": "query",
                },
                "other": {"description": "another param"},
            }
        }

    def test_model(self):
        api = Namespace("test")
        api.model("Person", {})
        assert "Person" in api.models
        assert isinstance(api.models["Person"], Model)

    def test_ordered_model(self):
        api = Namespace("test", ordered=True)
        api.model("Person", {})
        assert "Person" in api.models
        assert isinstance(api.models["Person"], OrderedModel)

    def test_schema_model(self):
        api = Namespace("test")
        api.schema_model("Person", {})
        assert "Person" in api.models

    def test_clone(self):
        api = Namespace("test")
        parent = api.model("Parent", {})
        api.clone("Child", parent, {})

        assert "Child" in api.models
        assert "Parent" in api.models

    def test_clone_with_multiple_parents(self):
        api = Namespace("test")
        grand_parent = api.model("GrandParent", {})
        parent = api.model("Parent", {})
        api.clone("Child", grand_parent, parent, {})

        assert "Child" in api.models
        assert "Parent" in api.models
        assert "GrandParent" in api.models

    def test_inherit(self):
        authorizations = {
            "apikey": {"type": "apiKey", "in": "header", "name": "X-API-KEY"}
        }
        api = Namespace("test", authorizations=authorizations)
        parent = api.model("Parent", {})
        api.inherit("Child", parent, {})

        assert "Parent" in api.models
        assert "Child" in api.models
        assert api.authorizations == authorizations

    def test_inherit_from_multiple_parents(self):
        api = Namespace("test")
        grand_parent = api.model("GrandParent", {})
        parent = api.model("Parent", {})
        api.inherit("Child", grand_parent, parent, {})

        assert "GrandParent" in api.models
        assert "Parent" in api.models
        assert "Child" in api.models

    def test_api_payload(self, app, client):
        api = restx.Api(app, validate=True)
        ns = restx.Namespace("apples")
        api.add_namespace(ns)

        fields = ns.model(
            "Person",
            {
                "name": restx.fields.String(required=True),
                "age": restx.fields.Integer,
                "birthdate": restx.fields.DateTime,
            },
        )

        @ns.route("/validation/")
        class Payload(restx.Resource):
            payload = None

            @ns.expect(fields)
            def post(self):
                Payload.payload = ns.payload
                return {}

        data = {
            "name": "John Doe",
            "age": 15,
        }

        client.post_json("/apples/validation/", data)

        assert Payload.payload == data

    def test_api_payload_strict_verification(self, app, client):
        api = restx.Api(app, validate=True)
        ns = restx.Namespace("apples")
        api.add_namespace(ns)

        fields = ns.model(
            "Person",
            {
                "name": restx.fields.String(required=True),
                "age": restx.fields.Integer,
                "birthdate": restx.fields.DateTime,
            },
            strict=True,
        )

        @ns.route("/validation/")
        class Payload(restx.Resource):
            payload = None

            @ns.expect(fields)
            def post(self):
                Payload.payload = ns.payload
                return {}

        data = {
            "name": "John Doe",
            "agge": 15,  # typo
        }

        resp = client.post_json("/apples/validation/", data, status=400)
        assert re.match(
            "Additional properties are not allowed \(u*'agge' was unexpected\)",
            resp["errors"][""],
        )



import flask_restx as restx


class Foo(restx.Resource):
    def get(self):
        return "data"


class ErrorsTest(object):
    def test_accept_default_application_json(self, app, client):
        api = restx.Api(app)
        api.add_resource(Foo, "/test/")

        res = client.get("/test/", headers={"Accept": None})
        assert res.status_code == 200
        assert res.content_type == "application/json"

    def test_accept_application_json_by_default(self, app, client):
        api = restx.Api(app)
        api.add_resource(Foo, "/test/")

        res = client.get("/test/", headers=[("Accept", "application/json")])
        assert res.status_code == 200
        assert res.content_type == "application/json"

    def test_accept_no_default_match_acceptable(self, app, client):
        api = restx.Api(app, default_mediatype=None)
        api.add_resource(Foo, "/test/")

        res = client.get("/test/", headers=[("Accept", "application/json")])
        assert res.status_code == 200
        assert res.content_type == "application/json"

    def test_accept_default_override_accept(self, app, client):
        api = restx.Api(app)
        api.add_resource(Foo, "/test/")

        res = client.get("/test/", headers=[("Accept", "text/plain")])
        assert res.status_code == 200
        assert res.content_type == "application/json"

    def test_accept_default_any_pick_first(self, app, client):
        api = restx.Api(app)

        @api.representation("text/plain")
        def text_rep(data, status_code, headers=None):
            resp = app.make_response((str(data), status_code, headers))
            return resp

        api.add_resource(Foo, "/test/")

        res = client.get("/test/", headers=[("Accept", "*/*")])
        assert res.status_code == 200
        assert res.content_type == "application/json"

    def test_accept_no_default_no_match_not_acceptable(self, app, client):
        api = restx.Api(app, default_mediatype=None)
        api.add_resource(Foo, "/test/")

        res = client.get("/test/", headers=[("Accept", "text/plain")])
        assert res.status_code == 406
        assert res.content_type == "application/json"

    def test_accept_no_default_custom_repr_match(self, app, client):
        api = restx.Api(app, default_mediatype=None)
        api.representations = {}

        @api.representation("text/plain")
        def text_rep(data, status_code, headers=None):
            resp = app.make_response((str(data), status_code, headers))
            return resp

        api.add_resource(Foo, "/test/")

        res = client.get("/test/", headers=[("Accept", "text/plain")])
        assert res.status_code == 200
        assert res.content_type == "text/plain"

    def test_accept_no_default_custom_repr_not_acceptable(self, app, client):
        api = restx.Api(app, default_mediatype=None)
        api.representations = {}

        @api.representation("text/plain")
        def text_rep(data, status_code, headers=None):
            resp = app.make_response((str(data), status_code, headers))
            return resp

        api.add_resource(Foo, "/test/")

        res = client.get("/test/", headers=[("Accept", "application/json")])
        assert res.status_code == 406
        assert res.content_type == "text/plain"

    def test_accept_no_default_match_q0_not_acceptable(self, app, client):
        """
        q=0 should be considered NotAcceptable,
        but this depends on werkzeug >= 1.0 which is not yet released
        so this test is expected to fail until we depend on werkzeug >= 1.0
        """
        api = restx.Api(app, default_mediatype=None)
        api.add_resource(Foo, "/test/")

        res = client.get("/test/", headers=[("Accept", "application/json; q=0")])
        assert res.status_code == 406
        assert res.content_type == "application/json"

    def test_accept_no_default_accept_highest_quality_of_two(self, app, client):
        api = restx.Api(app, default_mediatype=None)

        @api.representation("text/plain")
        def text_rep(data, status_code, headers=None):
            resp = app.make_response((str(data), status_code, headers))
            return resp

        api.add_resource(Foo, "/test/")

        res = client.get(
            "/test/", headers=[("Accept", "application/json; q=0.1, text/plain; q=1.0")]
        )
        assert res.status_code == 200
        assert res.content_type == "text/plain"

    def test_accept_no_default_accept_highest_quality_of_three(self, app, client):
        api = restx.Api(app, default_mediatype=None)

        @api.representation("text/html")
        @api.representation("text/plain")
        def text_rep(data, status_code, headers=None):
            resp = app.make_response((str(data), status_code, headers))
            return resp

        api.add_resource(Foo, "/test/")

        res = client.get(
            "/test/",
            headers=[
                (
                    "Accept",
                    "application/json; q=0.1, text/plain; q=0.3, text/html; q=0.2",
                )
            ],
        )
        assert res.status_code == 200
        assert res.content_type == "text/plain"

    def test_accept_no_default_no_representations(self, app, client):
        api = restx.Api(app, default_mediatype=None)
        api.representations = {}

        api.add_resource(Foo, "/test/")

        res = client.get("/test/", headers=[("Accept", "text/plain")])
        assert res.status_code == 406
        assert res.content_type == "text/plain"

    def test_accept_invalid_default_no_representations(self, app, client):
        api = restx.Api(app, default_mediatype="nonexistant/mediatype")
        api.representations = {}

        api.add_resource(Foo, "/test/")

        res = client.get("/test/", headers=[("Accept", "text/plain")])
        assert res.status_code == 500



import logging

import flask_restx as restx


class LoggingTest(object):
    def test_namespace_loggers_log_to_flask_app_logger(self, app, client, caplog):
        # capture Flask app logs
        caplog.set_level(logging.INFO, logger=app.logger.name)

        api = restx.Api(app)
        ns1 = api.namespace("ns1", path="/ns1")
        ns2 = api.namespace("ns2", path="/ns2")

        @ns1.route("/")
        class Ns1(restx.Resource):
            def get(self):
                ns1.logger.info("hello from ns1")
                pass

        @ns2.route("/")
        class Ns2(restx.Resource):
            def get(self):
                ns2.logger.info("hello from ns2")
                pass

        # debug log not shown
        client.get("/ns1/")
        matching = [r for r in caplog.records if r.message == "hello from ns1"]
        assert len(matching) == 1

        # info log shown
        client.get("/ns2/")
        matching = [r for r in caplog.records if r.message == "hello from ns2"]
        assert len(matching) == 1

    def test_defaults_to_app_level(self, app, client, caplog):
        caplog.set_level(logging.INFO, logger=app.logger.name)

        api = restx.Api(app)
        ns1 = api.namespace("ns1", path="/ns1")
        ns2 = api.namespace("ns2", path="/ns2")

        @ns1.route("/")
        class Ns1(restx.Resource):
            def get(self):
                ns1.logger.debug("hello from ns1")
                pass

        @ns2.route("/")
        class Ns2(restx.Resource):
            def get(self):
                ns2.logger.info("hello from ns2")
                pass

        # debug log not shown
        client.get("/ns1/")
        matching = [r for r in caplog.records if r.message == "hello from ns1"]
        assert len(matching) == 0

        # info log shown
        client.get("/ns2/")
        matching = [r for r in caplog.records if r.message == "hello from ns2"]
        assert len(matching) == 1

    def test_override_app_level(self, app, client, caplog):
        caplog.set_level(logging.DEBUG, logger=app.logger.name)

        api = restx.Api(app)
        ns1 = api.namespace("ns1", path="/ns1")
        ns1.logger.setLevel(logging.DEBUG)
        ns2 = api.namespace("ns2", path="/ns2")
        ns2.logger.setLevel(logging.INFO)

        @ns1.route("/")
        class Ns1(restx.Resource):
            def get(self):
                ns1.logger.debug("hello from ns1")
                pass

        @ns2.route("/")
        class Ns2(restx.Resource):
            def get(self):
                ns2.logger.debug("hello from ns2")
                pass

        # debug log shown from ns1
        client.get("/ns1/")
        matching = [r for r in caplog.records if r.message == "hello from ns1"]
        assert len(matching) == 1

        # debug not shown from ns2
        client.get("/ns2/")
        matching = [r for r in caplog.records if r.message == "hello from ns2"]
        assert len(matching) == 0

    def test_namespace_additional_handler(self, app, client, caplog, tmp_path):
        caplog.set_level(logging.INFO, logger=app.logger.name)
        log_file = tmp_path / "v1.log"

        api = restx.Api(app)
        ns1 = api.namespace("ns1", path="/ns1")
        # set up a file handler for ns1 only
        # FileHandler only supports Path object on Python >= 3.6 -> cast to str
        fh = logging.FileHandler(str(log_file))
        ns1.logger.addHandler(fh)

        ns2 = api.namespace("ns2", path="/ns2")

        @ns1.route("/")
        class Ns1(restx.Resource):
            def get(self):
                ns1.logger.info("hello from ns1")
                pass

        @ns2.route("/")
        class Ns2(restx.Resource):
            def get(self):
                ns2.logger.info("hello from ns2")
                pass

        # ns1 logs go to stdout and log_file
        client.get("/ns1/")
        matching = [r for r in caplog.records if r.message == "hello from ns1"]
        assert len(matching) == 1
        with log_file.open() as f:
            assert "hello from ns1" in f.read()

        # ns2 logs go to stdout only
        client.get("/ns2/")
        matching = [r for r in caplog.records if r.message == "hello from ns2"]
        assert len(matching) == 1
        with log_file.open() as f:
            assert "hello from ns1" in f.read()



import flask_restx as restx


class PayloadTest(object):
    def assert_errors(self, client, url, data, *errors):
        out = client.post_json(url, data, status=400)
        assert "message" in out
        assert "errors" in out
        for error in errors:
            assert error in out["errors"]

    def test_validation_false_on_constructor(self, app, client):
        api = restx.Api(app, validate=False)

        fields = api.model(
            "Person",
            {
                "name": restx.fields.String(required=True),
                "age": restx.fields.Integer,
                "birthdate": restx.fields.DateTime,
            },
        )

        @api.route("/validation/")
        class ValidationOff(restx.Resource):
            @api.expect(fields)
            def post(self):
                return {}

        data = client.post_json("/validation/", {})
        assert data == {}

    def test_validation_false_on_constructor_with_override(self, app, client):
        api = restx.Api(app, validate=False)

        fields = api.model(
            "Person",
            {
                "name": restx.fields.String(required=True),
                "age": restx.fields.Integer,
                "birthdate": restx.fields.DateTime,
            },
        )

        @api.route("/validation/")
        class ValidationOn(restx.Resource):
            @api.expect(fields, validate=True)
            def post(self):
                return {}

        self.assert_errors(client, "/validation/", {}, "name")

    def test_validation_true_on_constructor(self, app, client):
        api = restx.Api(app, validate=True)

        fields = api.model(
            "Person",
            {
                "name": restx.fields.String(required=True),
                "age": restx.fields.Integer,
                "birthdate": restx.fields.DateTime,
            },
        )

        @api.route("/validation/")
        class ValidationOff(restx.Resource):
            @api.expect(fields)
            def post(self):
                return {}

        self.assert_errors(client, "/validation/", {}, "name")

    def test_validation_true_on_constructor_with_override(self, app, client):
        api = restx.Api(app, validate=True)

        fields = api.model(
            "Person",
            {
                "name": restx.fields.String(required=True),
                "age": restx.fields.Integer,
                "birthdate": restx.fields.DateTime,
            },
        )

        @api.route("/validation/")
        class ValidationOff(restx.Resource):
            @api.expect(fields, validate=False)
            def post(self):
                return {}

        data = client.post_json("/validation/", {})
        assert data == {}

    def _setup_api_format_checker_tests(self, app, format_checker=None):
        class IPAddress(restx.fields.Raw):
            __schema_type__ = "string"
            __schema_format__ = "ipv4"

        api = restx.Api(app, format_checker=format_checker)
        model = api.model("MyModel", {"ip": IPAddress(required=True)})

        @api.route("/format_checker/")
        class TestResource(restx.Resource):
            @api.expect(model, validate=True)
            def post(self):
                return {}

    def test_format_checker_none_on_constructor(self, app, client):
        self._setup_api_format_checker_tests(app)

        out = client.post_json("/format_checker/", {"ip": "192.168.1"})
        assert out == {}

    def test_format_checker_object_on_constructor(self, app, client):
        from jsonschema import FormatChecker

        self._setup_api_format_checker_tests(app, format_checker=FormatChecker())

        out = client.post_json("/format_checker/", {"ip": "192.168.1"}, status=400)
        assert "ipv4" in out["errors"]["ip"]

    def test_validation_false_in_config(self, app, client):
        app.config["RESTX_VALIDATE"] = False
        api = restx.Api(app)

        fields = api.model(
            "Person",
            {
                "name": restx.fields.String(required=True),
                "age": restx.fields.Integer,
                "birthdate": restx.fields.DateTime,
            },
        )

        @api.route("/validation/")
        class ValidationOff(restx.Resource):
            @api.expect(fields)
            def post(self):
                return {}

        out = client.post_json("/validation/", {})

        # assert response.status_code == 200
        # out = json.loads(response.data.decode('utf8'))
        assert out == {}

    def test_validation_in_config(self, app, client):
        app.config["RESTX_VALIDATE"] = True
        api = restx.Api(app)

        fields = api.model(
            "Person",
            {
                "name": restx.fields.String(required=True),
                "age": restx.fields.Integer,
                "birthdate": restx.fields.DateTime,
            },
        )

        @api.route("/validation/")
        class ValidationOn(restx.Resource):
            @api.expect(fields)
            def post(self):
                return {}

        self.assert_errors(client, "/validation/", {}, "name")

    def test_api_payload(self, app, client):
        api = restx.Api(app, validate=True)

        fields = api.model(
            "Person",
            {
                "name": restx.fields.String(required=True),
                "age": restx.fields.Integer,
                "birthdate": restx.fields.DateTime,
            },
        )

        @api.route("/validation/")
        class Payload(restx.Resource):
            payload = None

            @api.expect(fields)
            def post(self):
                Payload.payload = api.payload
                return {}

        data = {
            "name": "John Doe",
            "age": 15,
        }

        client.post_json("/validation/", data)

        assert Payload.payload == data

    def test_validation_with_inheritance(self, app, client):
        """It should perform validation with inheritance (allOf/$ref)"""
        api = restx.Api(app, validate=True)

        fields = api.model(
            "Parent",
            {
                "name": restx.fields.String(required=True),
            },
        )

        child_fields = api.inherit(
            "Child",
            fields,
            {
                "age": restx.fields.Integer,
            },
        )

        @api.route("/validation/")
        class Inheritance(restx.Resource):
            @api.expect(child_fields)
            def post(self):
                return {}

        client.post_json(
            "/validation/",
            {
                "name": "John Doe",
                "age": 15,
            },
        )

        self.assert_errors(
            client,
            "/validation/",
            {
                "age": "15",
            },
            "name",
            "age",
        )

    def test_validation_on_list(self, app, client):
        """It should perform validation on lists"""
        api = restx.Api(app, validate=True)

        person = api.model(
            "Person",
            {
                "name": restx.fields.String(required=True),
                "age": restx.fields.Integer(required=True),
            },
        )

        family = api.model(
            "Family",
            {
                "name": restx.fields.String(required=True),
                "members": restx.fields.List(restx.fields.Nested(person)),
            },
        )

        @api.route("/validation/")
        class List(restx.Resource):
            @api.expect(family)
            def post(self):
                return {}

        self.assert_errors(
            client,
            "/validation/",
            {"name": "Doe", "members": [{"name": "Jonn"}, {"age": 42}]},
            "members.0.age",
            "members.1.name",
        )

    def _setup_expect_validation_single_resource_tests(self, app):
        # Setup a minimal Api with endpoint that expects in input payload
        # a single object of a resource
        api = restx.Api(app, validate=True)

        user = api.model("User", {"username": restx.fields.String()})

        @api.route("/validation/")
        class Users(restx.Resource):
            @api.expect(user)
            def post(self):
                return {}

    def _setup_expect_validation_collection_resource_tests(self, app):
        # Setup a minimal Api with endpoint that expects in input payload
        # one or more objects of a resource
        api = restx.Api(app, validate=True)

        user = api.model("User", {"username": restx.fields.String()})

        @api.route("/validation/")
        class Users(restx.Resource):
            @api.expect([user])
            def post(self):
                return {}

    def test_expect_validation_single_resource_success(self, app, client):
        self._setup_expect_validation_single_resource_tests(app)

        # Input payload is a valid JSON object
        out = client.post_json("/validation/", {"username": "alice"})
        assert {} == out

    def test_expect_validation_single_resource_error(self, app, client):
        self._setup_expect_validation_single_resource_tests(app)

        # Input payload is an invalid JSON object
        self.assert_errors(client, "/validation/", {"username": 123}, "username")

        # Input payload is a JSON array (expected JSON object)
        self.assert_errors(client, "/validation/", [{"username": 123}], "")

    def test_expect_validation_collection_resource_success(self, app, client):
        self._setup_expect_validation_collection_resource_tests(app)

        # Input payload is a valid JSON object
        out = client.post_json("/validation/", {"username": "alice"})
        assert {} == out

        # Input payload is a JSON array with valid JSON objects
        out = client.post_json(
            "/validation/", [{"username": "alice"}, {"username": "bob"}]
        )
        assert {} == out

    def test_expect_validation_collection_resource_error(self, app, client):
        self._setup_expect_validation_collection_resource_tests(app)

        # Input payload is an invalid JSON object
        self.assert_errors(client, "/validation/", {"username": 123}, "username")

        # Input payload is a JSON array but with an invalid JSON object
        self.assert_errors(
            client,
            "/validation/",
            [{"username": "alice"}, {"username": 123}],
            "username",
        )

    def test_validation_with_propagate(self, app, client):
        app.config["PROPAGATE_EXCEPTIONS"] = True
        api = restx.Api(app, validate=True)

        fields = api.model(
            "Person",
            {
                "name": restx.fields.String(required=True),
                "age": restx.fields.Integer,
                "birthdate": restx.fields.DateTime,
            },
        )

        @api.route("/validation/")
        class ValidationOff(restx.Resource):
            @api.expect(fields)
            def post(self):
                return {}

        self.assert_errors(client, "/validation/", {}, "name")

    def test_empty_payload(self, app, client):
        api = restx.Api(app, validate=True)

        @api.route("/empty/")
        class Payload(restx.Resource):
            def post(self):
                return {}

        response = client.post(
            "/empty/", data="", headers={"content-type": "application/json"}
        )

        assert response.status_code == 200
